# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import bz2
import concurrent.futures
import io
import os
import stat
import subprocess
import sys
import tarfile
import tempfile

import pytest
import rapidgzip
from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.tar import SQLiteIndexedTar, SQLiteIndexedTarUserData  # noqa: E402
from ratarmountcore.utils import RatarmountError  # noqa: E402


@pytest.mark.parametrize("parallelization", [1, 2, 4])
class TestSQLiteIndexedTarParallelized:
    @staticmethod
    def _create_file(tarArchive, name, contents):
        tinfo = tarfile.TarInfo(name)
        tinfo.size = len(contents)
        tarArchive.addfile(tinfo, io.BytesIO(contents.encode()))

    @staticmethod
    def _make_folder(tarArchive, name):
        tinfo = tarfile.TarInfo(name)
        tinfo.type = tarfile.DIRTYPE
        tarArchive.addfile(tinfo, io.BytesIO())

    @staticmethod
    def test_context_manager(parallelization):
        with copy_test_file("single-file.tar") as path, SQLiteIndexedTar(
            path, writeIndex=False, parallelization=parallelization
        ) as indexedTar:
            assert indexedTar.list('/')
            assert indexedTar.lookup('/')

            fileInfo = indexedTar.lookup('/bar')
            assert fileInfo
            assert fileInfo.size == 4
            userdata = fileInfo.userdata[-1]
            assert isinstance(userdata, SQLiteIndexedTarUserData)
            assert userdata.recursiondepth == 0
            assert not userdata.isgenerated

            assert not indexedTar.lookup('../')
            assert not indexedTar.lookup('../bar')

    @staticmethod
    def test_tar_bz2_with_parallelization(parallelization):
        with copy_test_file("2k-recursive-tars.tar.bz2") as path, SQLiteIndexedTar(
            path,
            clearIndexCache=True,
            recursive=False,
            parallelization=parallelization,
        ) as file:
            for folder in ['/', '/mimi']:
                assert file.lookup(folder)
                assert file.versions(folder) == 1
                assert file.list(folder)

            assert not file.list('/mimi/01995.tar')
            info = file.lookup('/mimi/01995.tar')
            assert info.userdata[0].offset == 21440512

            assert not file.list('/mimi/00105.tar')
            info = file.lookup('/mimi/00105.tar')
            assert info.userdata[0].offset == 1248256

    @staticmethod
    def test_recursive_tar_bz2_with_parallelization(parallelization):
        with copy_test_file("2k-recursive-tars.tar.bz2") as path, SQLiteIndexedTar(
            path,
            clearIndexCache=True,
            recursive=True,
            parallelization=parallelization,
        ) as file:
            assert file.list('/')
            assert file.list('/mimi')

            assert file.list('/mimi/01995.tar')
            info = file.lookup('/mimi/01995.tar/foo')
            assert info.userdata[0].offset == 21441024

            assert file.list('/mimi/00105.tar')
            info = file.lookup('/mimi/00105.tar/foo')
            assert info.userdata[0].offset == 1248768

    @staticmethod
    @pytest.mark.parametrize("recursive", [False, True])
    @pytest.mark.parametrize("maxRecursionDepth", [None, 0, 1, 2, 3, 4, 5])
    def test_deep_recursive(parallelization, recursive, maxRecursionDepth):
        with copy_test_file("packed-5-times.tar.gz") as path, SQLiteIndexedTar(
            path,
            clearIndexCache=True,
            recursive=recursive,
            recursionDepth=maxRecursionDepth,
            parallelization=parallelization,
        ) as mountSource:
            # packed-5-times.tar.gz -> /ufo_03.tar/ufo_02.tar/ufo_01.tar/ufo_00.tar/ufo
            maxDepth = 5  # 4 TAR archives + 1 compression layer
            recursionDepth = (maxDepth if recursive else 0) if maxRecursionDepth is None else maxRecursionDepth
            # Old default of undoing compression + TAR layer if nothing is specified.
            if not recursive and maxRecursionDepth is None:
                recursionDepth = 1

            assert mountSource.list('/')
            assert mountSource.lookup('/').userdata[-1] == SQLiteIndexedTarUserData(0, 0, False, False, True, 0)

            # Recursion depth:
            # file.tar -> / 0, /bar 0
            # nested-file.tar -> / 0, /bar.tar.version/1 0, /bar/ 1, /bar/foo 1
            # file.tar.gz -> / 0, /bar 1
            #    TODO: /file.tar 0  would be expected for completeness but it was not there in the past!
            # nested-file.tar -> / 0, /bar.tar.version/1 0, /bar/ 1, /bar/foo 1

            # The (generated) root has always recursion depth 0 because it will always be shown!
            assert mountSource.lookup('/').userdata[-1].recursiondepth == 0

            # Currently there is no way to get to this file when recursively mounted.
            # We would have to enable "/.versions/" to get older versions of the root folder or folders in general.
            if recursionDepth == 0:
                fileInfo = mountSource.lookup('/packed-5-times.tar')
                assert fileInfo
                userdata = fileInfo.userdata[-1]
                assert userdata.recursiondepth == 0
                assert stat.S_ISREG(fileInfo.mode)
                assert fileInfo.size == 51200

            # The recursion depth is 1 because of the gzip compression. Consider ufo_03.tar being inside
            # packaged-5-times.tar (without the .gz), then the recursion depth would have to be one less, i.e., 0!
            fileInfo = mountSource.lookup('/ufo_03.tar')
            if recursionDepth == 0:
                assert fileInfo is None
            elif recursionDepth == 1:
                assert fileInfo
                userdata = fileInfo.userdata[-1]
                assert userdata.recursiondepth == 1  # Not 0 because gzip compression counts as one layer of recursion.
                assert stat.S_ISREG(fileInfo.mode)
                assert not stat.S_ISDIR(fileInfo.mode)
                assert fileInfo.size == 40960
            else:
                assert fileInfo
                userdata = fileInfo.userdata[-1]
                assert userdata.recursiondepth == 2
                assert stat.S_ISDIR(fileInfo.mode)

                # Check that the older version is still reachable.
                fileInfo = mountSource.lookup('/ufo_03.tar', fileVersion=0 if recursionDepth == 0 else 1)
                assert fileInfo
                userdata = fileInfo.userdata[-1]
                assert userdata.recursiondepth == 1  # Not 0 because gzip compression counts as one layer of recursion.
                assert stat.S_ISREG(fileInfo.mode)
                assert not stat.S_ISDIR(fileInfo.mode)
                assert fileInfo.size == 40960

            def check_recursive_mount_point(path, depth):
                fileInfo = mountSource.lookup(path)
                if depth > recursionDepth:
                    return

                assert fileInfo
                userdata = fileInfo.userdata[-1]
                assert isinstance(userdata, SQLiteIndexedTarUserData)
                if depth <= recursionDepth:
                    assert stat.S_ISDIR(fileInfo.mode)
                    assert mountSource.list(path)
                else:
                    assert not stat.S_ISDIR(fileInfo.mode)

                assert userdata.recursiondepth == depth
                assert userdata.isgenerated
                assert userdata.istar

                if depth == maxDepth:
                    fileInfo = mountSource.lookup(path + '/ufo')
                    assert fileInfo
                    assert mountSource.open(fileInfo).read() == b'iriya\n'

            check_recursive_mount_point('/ufo_03.tar', 2)
            check_recursive_mount_point('/ufo_03.tar/ufo_02.tar', 3)
            check_recursive_mount_point('/ufo_03.tar/ufo_02.tar/ufo_01.tar', 4)
            check_recursive_mount_point('/ufo_03.tar/ufo_02.tar/ufo_01.tar/ufo_00.tar', 5)

    @staticmethod
    def test_compressed_tar(parallelization):
        with copy_test_file("packed-5-times.tar.gz") as path, SQLiteIndexedTar(
            path,
            clearIndexCache=True,
            parallelization=parallelization,
        ) as mountSource:
            assert mountSource.list('/')

            # See test_deep_recursive for recursion depth discussion.
            assert mountSource.lookup('/').userdata[-1] == SQLiteIndexedTarUserData(0, 0, False, False, True, 0)
            fileInfo = mountSource.lookup('/ufo_03.tar')
            assert fileInfo
            assert stat.S_ISREG(fileInfo.mode)
            assert fileInfo.userdata[-1].recursiondepth == 1

    @staticmethod
    def test_index_creation_and_loading(parallelization):
        with tempfile.TemporaryDirectory() as tmpDirectory:
            oldCurrentWorkingDirectory = os.getcwd()

            # Try with a writable directory and a non-writable current working directory
            # because FUSE also changes to root after mounting and forking to background.
            for directory in [tmpDirectory, '/']:
                os.chdir(directory)
                try:
                    archiveName = 'simple.bz2'
                    indexPath = 'simple.custom.index'
                    if directory != tmpDirectory:
                        archiveName = os.path.join(tmpDirectory, archiveName)
                        indexPath = os.path.join(tmpDirectory, indexPath)

                    contents = b"Hello World!"
                    with bz2.open(archiveName, "wb") as bz2File:
                        bz2File.write(contents)

                    def test_index(tarFileName, fileObject, indexFilePath, contents=contents):
                        TestSQLiteIndexedTarParallelized._test_index_creation_and_loading(
                            tarFileName, fileObject, indexFilePath, contents, parallelization
                        )

                    # States for arguments:
                    #  - file name: 3 (None, Path, ':memory:')
                    #  - archiveName: Optional[str]
                    #  - fileObject: Optional[IO]
                    # => 3*2*2 = 12 cases

                    with pytest.raises(RatarmountError):
                        test_index(None, None, ':memory:')
                    with pytest.raises(RatarmountError):
                        test_index(None, None, indexPath)
                    with pytest.raises(RatarmountError):
                        test_index(None, None, None)

                    test_index(archiveName, None, ':memory:')
                    test_index(archiveName, None, indexPath)
                    test_index(archiveName, None, None)

                    with open(archiveName, "rb") as file:
                        test_index("tarFileName", file, ':memory:')
                        test_index("tarFileName", file, indexPath)
                        test_index("tarFileName", file, None)

                        test_index(None, file, ':memory:')
                        test_index(None, file, indexPath)
                        test_index(None, file, None)

                finally:
                    os.chdir(oldCurrentWorkingDirectory)

    @staticmethod
    def _test_index_creation_and_loading(tarFileName, fileObject, indexFilePath, contents, parallelization):
        if indexFilePath:
            assert not os.path.exists(indexFilePath) or os.stat(indexFilePath).st_size == 0
        oldFolderContents = os.listdir('.')

        # Create index
        with SQLiteIndexedTar(
            fileObject=fileObject,
            tarFileName=tarFileName,
            writeIndex=True,
            clearIndexCache=True,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
        ):
            pass

        # When opening a file object without a path or with path ':memory:',
        # no index should have been created anywhere!
        if (fileObject is not None and indexFilePath is None) or indexFilePath == ':memory:':
            assert oldFolderContents == os.listdir('.')
            return

        createdIndexFilePath = indexFilePath
        if not indexFilePath and not fileObject and tarFileName:
            createdIndexFilePath = tarFileName + '.index.sqlite'
        assert os.stat(createdIndexFilePath).st_size > 0

        # Read from index
        indexedFile = SQLiteIndexedTar(
            fileObject=fileObject,
            tarFileName=tarFileName,
            writeIndex=False,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        )

        objectName = '<file object>' if tarFileName is None else tarFileName
        expectedName = os.path.basename(tarFileName).rsplit('.', 1)[0] if fileObject is None else objectName

        folderList = indexedFile.list("/")
        assert isinstance(folderList, dict)
        if isinstance(folderList, dict):
            # https://github.com/PyCQA/pylint/issues/1162
            assert expectedName in folderList  # pylint: disable=unsupported-membership-test
            assert folderList[expectedName].size == len(contents)  # pylint: disable=unsubscriptable-object

        finfo = indexedFile.lookup("/" + expectedName)
        assert finfo.size == len(contents)
        assert indexedFile.read(finfo, size=len(contents), offset=0) == contents
        assert indexedFile.read(finfo, size=3, offset=3) == contents[3:6]

        if createdIndexFilePath:
            os.remove(createdIndexFilePath)

    @staticmethod
    def test_list_and_versions(parallelization):
        with tempfile.NamedTemporaryFile(suffix=".tar.gz") as tmpTarFile:
            with tarfile.open(name=tmpTarFile.name, mode="w:gz") as tarFile:
                create_file = TestSQLiteIndexedTarParallelized._create_file
                make_folder = TestSQLiteIndexedTarParallelized._make_folder

                create_file(tarFile, "./README.md", "hello world")
                make_folder(tarFile, "./src")
                create_file(tarFile, "./src/test.sh", "echo hi")
                make_folder(tarFile, "./dist")
                make_folder(tarFile, "./dist/a")
                make_folder(tarFile, "./dist/a/b")
                create_file(tarFile, "./dist/a/b/test2.sh", "echo two")

            with SQLiteIndexedTar(tmpTarFile.name, clearIndexCache=True, parallelization=parallelization) as indexedTar:
                folders = []
                files = []

                foldersToRecurse = ["/"]
                while foldersToRecurse:
                    folder = foldersToRecurse.pop()
                    folderContents = indexedTar.list(folder)
                    assert isinstance(folderContents, dict)
                    for name in folderContents:  # pylint: disable=not-an-iterable
                        path = os.path.join(folder, name)
                        fileInfo = indexedTar.lookup(path)
                        if not fileInfo:
                            continue

                        if stat.S_ISDIR(fileInfo.mode):
                            folders.append(path)
                            foldersToRecurse.append(path)
                        else:
                            files.append(path)

                assert set(folders) == {"/dist", "/dist/a", "/dist/a/b", "/src"}
                assert set(files) == {"/dist/a/b/test2.sh", "/src/test.sh", "/README.md"}

                for path in folders + files:
                    assert indexedTar.versions(path) == 1

    @staticmethod
    def test_open(parallelization):
        with tempfile.NamedTemporaryFile(suffix=".tar.gz") as tmpTarFile:
            repeatCount = 10000

            with tarfile.open(name=tmpTarFile.name, mode="w:gz") as tarFile:
                create_file = TestSQLiteIndexedTarParallelized._create_file
                create_file(tarFile, "increasing.dat", "".join(["0123456789"] * repeatCount))
                create_file(tarFile, "decreasing.dat", "".join(["9876543210"] * repeatCount))

            with SQLiteIndexedTar(tmpTarFile.name, clearIndexCache=True, parallelization=parallelization) as indexedTar:
                iFile = indexedTar.open(indexedTar.lookup("/increasing.dat"))
                dFile = indexedTar.open(indexedTar.lookup("/decreasing.dat"))

                for i in range(repeatCount):
                    try:
                        assert iFile.read(10) == b"0123456789"
                        assert dFile.read(10) == b"9876543210"
                    except AssertionError as e:
                        print("Reading failed in iteration:", i)
                        raise e

    @staticmethod
    def test_multithreaded_reading(parallelization):
        parallelism = parallelization * 6  # Need a bit more parallelism to trigger bugs easier
        with tempfile.NamedTemporaryFile(suffix=".tar.gz") as tmpTarFile, concurrent.futures.ThreadPoolExecutor(
            parallelism
        ) as pool:
            repeatCount = 10000

            with tarfile.open(name=tmpTarFile.name, mode="w:gz") as tarFile:
                create_file = TestSQLiteIndexedTarParallelized._create_file
                create_file(tarFile, "increasing.dat", "".join(["0123456789"] * repeatCount))
                create_file(tarFile, "decreasing.dat", "".join(["9876543210"] * repeatCount))

            with SQLiteIndexedTar(tmpTarFile.name, clearIndexCache=True, parallelization=parallelization) as indexedTar:

                def read_sequences(file, useIncreasing):
                    for i in range(repeatCount):
                        try:
                            if useIncreasing:
                                assert file.read(10) == b"0123456789"
                            else:
                                assert file.read(10) == b"9876543210"
                        except AssertionError as e:
                            print("Reading failed in iteration:", i)
                            raise e
                    return True

                files = [
                    indexedTar.open(indexedTar.lookup(f"/{'in' if i % 2 == 0 else 'de'}creasing.dat"))
                    for i in range(parallelism)
                ]

                results = [pool.submit(read_sequences, files[i], i % 2 == 0) for i in range(parallelism)]
                for result in results:
                    result.result()

                for file in files:
                    file.close()

    @staticmethod
    def test_appending_to_small_archive(parallelization, tmpdir):
        create_file = TestSQLiteIndexedTarParallelized._create_file
        make_folder = TestSQLiteIndexedTarParallelized._make_folder

        # Create a simple small TAR
        tarPath = os.path.join(tmpdir, "foo.tar")
        with tarfile.open(name=tarPath, mode="w:") as tarFile:
            create_file(tarFile, "foo", "bar")

        # Create index
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=True,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert not indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/foo")
            assert not indexedTar.exists("/bar")
            assert not indexedTar.exists("/folder")

        # Append small file to TAR
        with tarfile.open(name=tarPath, mode="a:") as tarFile:
            create_file(tarFile, "bar", "foo")

        # Create index but only go over new files
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert not indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/foo")
            assert indexedTar.exists("/bar")
            assert not indexedTar.exists("/folder")

        # Append empty folder to TAR
        with tarfile.open(name=tarPath, mode="a:") as tarFile:
            make_folder(tarFile, "folder")

        # Create index but only go over new files
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert not indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/foo")
            assert indexedTar.exists("/bar")
            assert indexedTar.exists("/folder")

        # Append a sparse file
        sparsePath = os.path.join(tmpdir, "sparse")
        with open(sparsePath, "wb"):
            pass
        os.truncate(sparsePath, 1024 * 1024)
        # The tarfile module only has read support for sparse files, therefore use GNU tar to do it
        subprocess.run(["tar", "--append", "-f", tarPath, sparsePath], check=True)

        # Create index. Because of the sparse file at the end, it might be recreated from scratch.
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert not indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/foo")
            assert indexedTar.exists("/bar")
            assert indexedTar.exists("/folder")

        # Append small file to TAR after sparse file!
        with tarfile.open(name=tarPath, mode="a:") as tarFile:
            create_file(tarFile, "bar2", "foo")

        # Create index but only go over new files
        print("=== Update Index With New File After Sparse File ===")
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert not indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/foo")
            assert indexedTar.exists("/bar")
            assert indexedTar.exists("/folder")
            assert indexedTar.exists("/bar2")

    @staticmethod
    def test_appending_to_large_archive(parallelization, tmpdir):
        create_file = TestSQLiteIndexedTarParallelized._create_file
        make_folder = TestSQLiteIndexedTarParallelized._make_folder

        # Create a TAR large in size as well as file count
        tarPath = os.path.join(tmpdir, "foo.tar")
        with copy_test_file(
            "tar-with-300-folders-with-1000-files-0B-files.tar.bz2"
        ) as path, rapidgzip.IndexedBzip2File(path) as file, open(tarPath, 'wb') as extracted:
            while True:
                data = file.read(1024 * 1024)
                if not data:
                    break
                extracted.write(data)

        # Create index
        print("\n=== Create Index ===")
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=True,
            indexFilePath=indexFilePath,
            parallelization=os.cpu_count(),
            printDebug=3,
        ) as indexedTar:
            assert not indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/00000000000000000000000000000282/00000000000000000000000000000976")
            assert not indexedTar.exists("/bar")
            assert not indexedTar.exists("/folder")

        # Append small file to TAR
        with tarfile.open(name=tarPath, mode="a:") as tarFile:
            create_file(tarFile, "bar", "foo")

        # Create index but only go over new files
        print("\n=== Update Index With New File ===")
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/00000000000000000000000000000282/00000000000000000000000000000976")
            assert indexedTar.exists("/bar")
            assert not indexedTar.exists("/folder")

        # Append empty folder to TAR
        with tarfile.open(name=tarPath, mode="a:") as tarFile:
            make_folder(tarFile, "folder")

        # Create index but only go over new files
        print("\n=== Update Index With New Folder ===")
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/00000000000000000000000000000282/00000000000000000000000000000976")
            assert indexedTar.exists("/bar")
            assert indexedTar.exists("/folder")

        # Append a sparse file
        sparsePath = os.path.join(tmpdir, "sparse")
        with open(sparsePath, "wb"):
            pass
        os.truncate(sparsePath, 1024 * 1024)
        # The tarfile module only has read support for sparse files, therefore use GNU tar to do it
        subprocess.run(["tar", "--append", "-f", tarPath, sparsePath], check=True)

        # Create index. Because of the sparse file at the end, it might be recreated from scratch.
        print("\n=== Update Index With Sparse File ===")
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
            printDebug=3,
        ) as indexedTar:
            assert indexedTar.hasBeenAppendedTo
            assert indexedTar.exists("/00000000000000000000000000000282/00000000000000000000000000000976")
            assert indexedTar.exists("/bar")
            assert indexedTar.exists("/folder")

        # Append small file to TAR after sparse file!
        with tarfile.open(name=tarPath, mode="a:") as tarFile:
            create_file(tarFile, "bar2", "foo")

        # Create index but only go over new files
        print("\n=== Update Index With New File After Sparse File ===")
        indexFilePath = os.path.join(tmpdir, "foo.tar.index")
        with SQLiteIndexedTar(
            tarFileName=tarPath,
            writeIndex=True,
            clearIndexCache=False,
            indexFilePath=indexFilePath,
            parallelization=parallelization,
        ) as indexedTar:
            # assert indexedTar.hasBeenAppendedTo  # TODO
            assert indexedTar.exists("/00000000000000000000000000000282/00000000000000000000000000000976")
            assert indexedTar.exists("/bar")
            assert indexedTar.exists("/folder")
            assert indexedTar.exists("/bar2")



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import bz2
import os
import sys
from pathlib import Path

import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.factory import open_mount_source  # noqa: E402
from ratarmountcore.mountsource.formats.tar import SQLiteIndexedTar  # noqa: E402
from ratarmountcore.mountsource.formats.zip import ZipMountSource  # noqa: E402


def find_test_file(relativePathOrName):
    for i in range(3):
        path = os.path.sep.join([".."] * i + ["tests", relativePathOrName])
        if os.path.exists(path):
            return path
    return relativePathOrName


@pytest.mark.parametrize("transform_path", [str, Path])
class TestOpenMountSource:
    @staticmethod
    def test_joining_archive(tmp_path, transform_path):
        compressed = bz2.compress(b"foobar")
        (tmp_path / "foo.001").write_bytes(compressed[: len(compressed) // 2])
        (tmp_path / "foo.002").write_bytes(compressed[len(compressed) // 2 :])

        for path in [tmp_path / "foo.001", tmp_path / "foo.002"]:
            with open_mount_source(transform_path(path)) as mountSource:
                fileInfo = mountSource.lookup("/<file object>")
                assert fileInfo
                assert mountSource.open(fileInfo).read() == b"foobar"

    @staticmethod
    def test_joining_file(tmp_path, transform_path):
        (tmp_path / "foo.001").write_bytes(b"foo")
        (tmp_path / "foo.002").write_bytes(b"bar")

        for path in [tmp_path / "foo.001", tmp_path / "foo.002"]:
            with open_mount_source(transform_path(path)) as mountSource:
                fileInfo = mountSource.lookup("/foo")
                assert fileInfo
                assert mountSource.open(fileInfo).read() == b"foobar"

    @staticmethod
    def test_joining_files_exceeding_handle_limit(tmp_path, transform_path):
        result = b''
        for i in range(1100):  # Default on my system is 1024
            (tmp_path / f"foo.{i:03}").write_bytes(str(i).encode())
            result += str(i).encode()

        with open_mount_source(transform_path(tmp_path / "foo.005")) as mountSource:
            fileInfo = mountSource.lookup("/foo")
            assert fileInfo
            assert mountSource.open(fileInfo).read() == result

    @staticmethod
    def test_chimera_file(transform_path):
        chimeraFilePath = transform_path(Path(find_test_file("chimera-tbz2-zip")))
        indexPath = Path(str(chimeraFilePath) + ".index.sqlite")
        if indexPath.exists():
            indexPath.unlink()

        # Check simple open and that index files are NOT created because they are too small.
        with open_mount_source(
            chimeraFilePath, writeIndex=True, prioritizedBackends=['zipfile', 'rapidgzip']
        ) as mountSource:
            assert isinstance(mountSource, ZipMountSource)
            files = mountSource.list("/")
            assert files

            assert not indexPath.exists()

        # Same as above, but force index creation by lowering the file count threshold.
        with open_mount_source(
            chimeraFilePath,
            writeIndex=True,
            prioritizedBackends=['zipfile', 'rapidgzip'],
            indexMinimumFileCount=0,
        ) as mountSource:
            assert isinstance(mountSource, ZipMountSource)
            files = mountSource.list("/")
            assert files

            fileInfo = mountSource.lookup("/foo/fighter/ufo")
            assert fileInfo
            with mountSource.open(fileInfo) as file:
                assert file.read() == b"iriya\n"

            assert indexPath.exists()

        # Check that everything works fine even if the index exists and the backend order is reversed.
        #
        # "Works fine" refers to a valid file hierarchy being shown and that files can be read without errors.
        # I'm not decided whether the next backend should be tried or whether the index should be recreated
        # by the first backend after an inconsistency has been noticed. The latter is easier to implement
        # and more consistent. I think, only after implementing storing the backend name into the index,
        # should the next backend be tried instead of it being overwritten and recreated.
        assert indexPath.exists()
        with open_mount_source(
            chimeraFilePath,
            writeIndex=True,
            prioritizedBackends=['rapidgzip', 'zipfile'],
            indexMinimumFileCount=0,
        ) as mountSource:
            assert isinstance(mountSource, ZipMountSource)
            files = mountSource.list("/")
            assert files

            fileInfo = mountSource.lookup("/foo/fighter/ufo")
            assert fileInfo
            with mountSource.open(fileInfo) as file:
                assert file.read() == b"iriya\n"

            assert indexPath.exists()

        indexPath.unlink()

        # Index file is always created for compressed files such as .tar.bz2
        with open_mount_source(
            chimeraFilePath, writeIndex=True, prioritizedBackends=['rapidgzip', 'zipfile']
        ) as mountSource:
            assert isinstance(mountSource, SQLiteIndexedTar)
            files = mountSource.list("/")
            assert files

            fileInfo = mountSource.lookup("/bar")
            assert fileInfo
            with mountSource.open(fileInfo) as file:
                assert file.read() == b"foo\n"

            assert indexPath.exists()

        # Check that everything works fine even if the index exists and the backend order is reversed.
        assert indexPath.exists()
        with open_mount_source(
            chimeraFilePath,
            writeIndex=True,
            prioritizedBackends=['zipfile', 'rapidgzip'],
            indexMinimumFileCount=0,
        ) as mountSource:
            assert isinstance(mountSource, SQLiteIndexedTar)
            files = mountSource.list("/")
            print("files:", files)
            assert files

            fileInfo = mountSource.lookup("/bar")
            assert fileInfo
            with mountSource.open(fileInfo) as file:
                assert file.read() == b"foo\n"

            assert indexPath.exists()

        indexPath.unlink()



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import concurrent.futures
import io
import os
import sys
import tempfile
import threading

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import pytest  # noqa: E402
from ratarmountcore.StenciledFile import (  # noqa: E402
    JoinedFile,
    JoinedFileFromFactory,
    RawStenciledFile,
    StenciledFile,
)

testData = b"1234567890"
tmpFile = tempfile.TemporaryFile()
tmpFile.write(testData)


randomTestData = os.urandom(128 * 1024)
randomTmpFile = tempfile.TemporaryFile()
randomTmpFile.write(randomTestData)


class TestStenciledFile:
    @staticmethod
    def _create_stenciled_file(file, stencils):
        return StenciledFile(fileStencils=[(file, *stencil) for stencil in stencils])

    @staticmethod
    def test_empty_file():
        assert StenciledFile([(tmpFile, 0, 0)]).read() == b""

        with StenciledFile([]) as file:
            assert file.readable()
            assert file.seekable()
            assert file.read() == b""
            assert file.seek(0, io.SEEK_SET) == 0
            assert file.seek(0, io.SEEK_CUR) == 0
            assert file.seek(0, io.SEEK_END) == 0
            assert file.read() == b""

    @staticmethod
    def test_find_stencil():
        stenciledFile = RawStenciledFile(
            [(tmpFile, *stencil) for stencil in [(1, 2), (2, 2), (0, 2), (4, 4), (1, 8), (0, 1)]]
        )
        expectedResults = [0, 0, 1, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5]
        for offset, iExpectedStencil in enumerate(expectedResults):
            assert stenciledFile._find_stencil(offset) == iExpectedStencil

    @staticmethod
    def test_single_stencil():
        assert StenciledFile([(tmpFile, 0, 1)]).read() == b"1"
        assert StenciledFile([(tmpFile, 0, 2)]).read() == b"12"
        assert StenciledFile([(tmpFile, 0, 3)]).read() == b"123"
        assert StenciledFile([(tmpFile, 0, len(testData))]).read() == testData

    @staticmethod
    def test_1_byte_stencils():
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(0, 1), (1, 1)]).read() == b"12"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(0, 1), (2, 1)]).read() == b"13"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 1), (0, 1)]).read() == b"21"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(0, 1), (1, 1), (2, 1)]).read() == b"123"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 1), (2, 1), (0, 1)]).read() == b"231"

    @staticmethod
    def test_2_byte_stencils():
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(0, 2), (1, 2)]).read() == b"1223"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(0, 2), (2, 2)]).read() == b"1234"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (0, 2)]).read() == b"2312"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(0, 2), (1, 2), (2, 2)]).read() == b"122334"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read() == b"233412"

    @staticmethod
    def test_read_with_size():
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(0) == b""
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(1) == b"2"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(2) == b"23"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(3) == b"233"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(4) == b"2334"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(5) == b"23341"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(6) == b"233412"
        assert TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)]).read(7) == b"233412"

    @staticmethod
    def test_seek_and_tell():
        file = TestStenciledFile._create_stenciled_file(tmpFile, [(1, 2), (2, 2), (0, 2)])

        assert file.readable()
        assert file.seekable()
        assert not file.writable()
        assert not file.closed

        for i in range(7):
            assert file.tell() == i
            file.read(1)
        for i in reversed(range(6)):
            assert file.seek(-1, io.SEEK_CUR) == i
            assert file.tell() == i
        assert file.seek(0, io.SEEK_END) == 6
        assert file.tell() == 6
        assert file.seek(20, io.SEEK_END) == 26
        assert file.tell() == 26
        assert file.read(1) == b""
        assert file.seek(-6, io.SEEK_END) == 0
        assert file.read(1) == b"2"

        with pytest.raises(io.UnsupportedOperation):
            file.fileno()

        file.close()
        assert file.closed

    @staticmethod
    def test_reading_from_shared_file():
        file1 = StenciledFile([(tmpFile, 0, len(testData))])
        file2 = StenciledFile([(tmpFile, 0, len(testData))])
        for i in range(len(testData)):
            assert file1.read(1) == testData[i : i + 1]
            assert file2.read(1) == testData[i : i + 1]

    @staticmethod
    def successive_reads_test(lock=None):
        file = StenciledFile([(randomTmpFile, 0, len(randomTestData))], lock)
        batchSize = 1024
        for i in range(len(randomTestData) // batchSize):
            assert file.read(batchSize) == randomTestData[i * batchSize : (i + 1) * batchSize]

    @staticmethod
    def test_successive_reads():
        TestStenciledFile.successive_reads_test()

    @staticmethod
    def test_multithreaded_reading():
        parallelism = 24
        with concurrent.futures.ThreadPoolExecutor(24) as pool:
            lock = threading.Lock()
            results = [pool.submit(TestStenciledFile.successive_reads_test, lock) for _ in range(parallelism)]
            for result in results:
                result.result()


class TestJoinedFile:
    @staticmethod
    def test_empty_file():
        assert JoinedFile([io.BytesIO(b"")]).read() == b""

    @staticmethod
    def test_single_file():
        assert JoinedFile([io.BytesIO(b"f")]).read() == b"f"
        assert JoinedFile([io.BytesIO(b"fo")]).read() == b"fo"
        assert JoinedFile([io.BytesIO(b"foo")]).read() == b"foo"

    @staticmethod
    def test_single_file_non_complete_read():
        assert JoinedFile([io.BytesIO(b"f")]).read(1) == b"f"
        assert JoinedFile([io.BytesIO(b"fo")]).read(1) == b"f"
        assert JoinedFile([io.BytesIO(b"foo")]).read(1) == b"f"

    @staticmethod
    def test_single_file_seek_read():
        file = JoinedFile([io.BytesIO(b"foobar")])

        assert file.readable()
        assert file.seekable()
        assert not file.writable()
        assert not file.closed

        assert file.seek(1) == 1
        assert file.read(1) == b"o"
        assert file.read() == b"obar"
        assert file.seek(-1, io.SEEK_END) == 5
        assert file.read() == b"r"

        with pytest.raises(io.UnsupportedOperation):
            file.fileno()

        file.close()
        assert file.closed

    @staticmethod
    def test_two_files_full_read():
        assert JoinedFile([io.BytesIO(b""), io.BytesIO(b"")]).read() == b""
        assert JoinedFile([io.BytesIO(b""), io.BytesIO(b"foo")]).read() == b"foo"
        assert JoinedFile([io.BytesIO(b"foo"), io.BytesIO(b"")]).read() == b"foo"
        assert JoinedFile([io.BytesIO(b"bar"), io.BytesIO(b"foo")]).read() == b"barfoo"

    @staticmethod
    def test_two_files_seak_and_read():
        file = JoinedFile([io.BytesIO(b"bar"), io.BytesIO(b"foo")])
        assert file.read(1) == b"b"
        assert file.tell() == 1

        assert file.seek(3) == 3
        assert file.tell() == 3

        assert file.read(2) == b"fo"
        assert file.tell() == 5

        assert file.seek(-4, io.SEEK_END) == 2
        assert file.tell() == 2

        assert file.read() == b"rfoo"
        assert file.tell() == 6


class TestJoinedFileFromFactory:
    @staticmethod
    def test_empty_file():
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"")]).read() == b""

    @staticmethod
    def test_single_file():
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"f")]).read() == b"f"
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"fo")]).read() == b"fo"
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"foo")]).read() == b"foo"

    @staticmethod
    def test_single_file_non_complete_read():
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"f")]).read(1) == b"f"
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"fo")]).read(1) == b"f"
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"foo")]).read(1) == b"f"

    @staticmethod
    def test_single_file_seek_read():
        file = JoinedFileFromFactory([lambda: io.BytesIO(b"foobar")])

        assert file.readable()
        assert file.seekable()
        assert not file.writable()
        assert not file.closed

        assert file.seek(1) == 1
        assert file.read(1) == b"o"
        assert file.read() == b"obar"
        assert file.seek(-1, io.SEEK_END) == 5
        assert file.read() == b"r"

        with pytest.raises(io.UnsupportedOperation):
            file.fileno()

        file.close()
        assert file.closed

    @staticmethod
    def test_two_files_full_read():
        assert JoinedFileFromFactory([lambda: io.BytesIO(b""), lambda: io.BytesIO(b"")]).read() == b""
        assert JoinedFileFromFactory([lambda: io.BytesIO(b""), lambda: io.BytesIO(b"foo")]).read() == b"foo"
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"foo"), lambda: io.BytesIO(b"")]).read() == b"foo"
        assert JoinedFileFromFactory([lambda: io.BytesIO(b"bar"), lambda: io.BytesIO(b"foo")]).read() == b"barfoo"

    @staticmethod
    def test_two_files_seak_and_read():
        file = JoinedFileFromFactory([lambda: io.BytesIO(b"bar"), lambda: io.BytesIO(b"foo")])
        assert file.read(1) == b"b"
        assert file.tell() == 1

        assert file.seek(3) == 3
        assert file.tell() == 3

        assert file.read(2) == b"fo"
        assert file.tell() == 5

        assert file.seek(-4, io.SEEK_END) == 2
        assert file.tell() == 2

        assert file.read() == b"rfoo"
        assert file.tell() == 6

    @staticmethod
    def test_joining_files(tmp_path):
        files = [tmp_path / name for name in ["foo.001", "foo.002"]]
        files[0].write_bytes(b"foo")
        files[1].write_bytes(b"bar")

        factories = [lambda file=file: open(file, 'rb') for file in files]
        assert len(factories) == 2
        assert factories[0]().read() == b"foo"
        assert factories[1]().read() == b"bar"

        assert JoinedFileFromFactory(factories).read() == b"foobar"



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import io
import os
import stat
import sys
import tarfile
import tempfile
import time
from pathlib import Path

import pytest
from helpers import copy_test_file, find_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.libarchive import IterableArchive, LibarchiveMountSource  # noqa: E402


class TestLibarchiveMountSource:
    @staticmethod
    @pytest.mark.parametrize('compression', ['7z', 'rar', 'zip'])
    def test_simple_usage(compression):
        with copy_test_file('folder-symlink.' + compression) as path, LibarchiveMountSource(path) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'

            # Links are not resolved by the mount source but by FUSE, i.e., descending into a link to a folder
            # will not work. This behavior may change in the future.
            for linkPath in ['/foo/jet']:
                assert mountSource.lookup(linkPath)
                assert mountSource.versions(linkPath) == 1
                assert not mountSource.list(linkPath)
                fileInfo = mountSource.lookup(linkPath)
                assert fileInfo.linkname == 'fighter'

    @staticmethod
    @pytest.mark.parametrize('compression', ['7z', 'rar', 'zip'])
    def test_transform(compression):
        with copy_test_file('folder-symlink.' + compression) as path, LibarchiveMountSource(
            path, transform=("(.)/(.)", r"\1_\2")
        ) as mountSource:
            for folder in ['/', '/foo', '/foo_fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)
                assert mountSource.versions(folder) == 1

            for filePath in ['/foo_fighter_ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'

    @staticmethod
    # 7z : libarchive.exception.ArchiveError: The file content is encrypted, but currently not supported
    #      (errno=-1, retcode=-30, archive_p=94443813387248)
    # RAR: libarchive.exception.ArchiveError: Unsupported block header size (was 4, max is 2)
    #      (errno=84, retcode=-30, archive_p=94443813892640)
    # Basically only ZIP has encryption support provided by libarchive, much less than I would have thought.
    # https://github.com/libarchive/libarchive/issues/579#issuecomment-118440525
    # @pytest.mark.parametrize("compression", ["7z", "rar", "zip"])
    @pytest.mark.parametrize('compression', ['zip'])
    def test_password(compression):
        with copy_test_file('encrypted-nested-tar.' + compression) as path, LibarchiveMountSource(
            path, passwords=['foo']
        ) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'

    @staticmethod
    @pytest.mark.parametrize('compression', ['bz2', 'gz', 'lrz', 'lz4', 'lzip', 'lzma', 'lzo', 'xz', 'Z', 'zst'])
    def test_stream_compressed(compression):
        with copy_test_file('simple.' + compression) as path, LibarchiveMountSource(
            path, passwords=['foo']
        ) as mountSource:
            for folder in ['/']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/simple']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'foo fighter\n'
                    assert file.seek(4) == 4
                    assert file.read() == b'fighter\n'

    @staticmethod
    @pytest.mark.parametrize(
        ('path', 'lineSize'),
        [
            # libarchive bug: https://github.com/libarchive/libarchive/issues/2106
            # ('two-large-files-32Ki-lines-each-1024B.7z', 1024),
            ('two-large-files-32Ki-lines-each-1023B.7z', 1023),
        ],
    )
    def test_file_independence(path, lineSize):
        with copy_test_file(path) as copiedPath, LibarchiveMountSource(copiedPath) as mountSource:
            with mountSource.open(mountSource.lookup('zeros-32-MiB.txt')) as fileWithZeros:
                expectedZeros = b'0' * (lineSize - 1) + b'\n'
                assert fileWithZeros.read(lineSize) == expectedZeros
                assert fileWithZeros.tell() == lineSize
                assert fileWithZeros.seek(-lineSize, io.SEEK_END)
                actualZeros = fileWithZeros.read()
                assert len(actualZeros) == len(expectedZeros)
                assert actualZeros == expectedZeros

            with mountSource.open(mountSource.lookup('zeros-32-MiB.txt')) as fileWithZeros, mountSource.open(
                mountSource.lookup('spaces-32-MiB.txt')
            ) as fileWithSpaces:
                expectedSpaces = b' ' * (lineSize - 1) + b'\n'
                expectedZeros = b'0' * (lineSize - 1) + b'\n'

                assert fileWithSpaces.read(lineSize) == expectedSpaces
                assert fileWithSpaces.tell() == lineSize

                assert fileWithZeros.read(lineSize) == expectedZeros
                assert fileWithZeros.tell() == lineSize

                assert fileWithSpaces.tell() == lineSize

                assert fileWithSpaces.seek(-lineSize, io.SEEK_END)
                assert fileWithZeros.seek(-lineSize, io.SEEK_END)
                assert fileWithSpaces.read() == expectedSpaces
                assert fileWithZeros.read() == expectedZeros

                # Seek backwards inside of buffer

                assert fileWithSpaces.seek(-lineSize + 1, io.SEEK_END)
                assert fileWithZeros.seek(-lineSize + 1, io.SEEK_END)
                assert fileWithSpaces.read() == expectedSpaces[1:]
                assert fileWithZeros.read() == expectedZeros[1:]

                # Seek backwards outside of buffer

                assert fileWithSpaces.seek(1) == 1
                assert fileWithSpaces.read(lineSize - 1) == expectedSpaces[1:]
                assert fileWithSpaces.tell() == lineSize

                assert fileWithZeros.seek(1) == 1
                assert fileWithZeros.read(lineSize - 1) == expectedZeros[1:]
                assert fileWithZeros.tell() == lineSize

    @staticmethod
    def test_file_object_reader():
        memoryFile = io.BytesIO(Path(find_test_file('folder-symlink.zip')).read_bytes())
        with IterableArchive(memoryFile) as archive:
            while True:
                entry = archive.next_entry()
                if entry is None:
                    break
                fileInfo = entry.convert_to_row(0, lambda x: x)
                assert fileInfo

    @staticmethod
    def _create_file(tarArchive, name, contents):
        tinfo = tarfile.TarInfo(name)
        tinfo.size = len(contents)
        tarArchive.addfile(tinfo, io.BytesIO(contents.encode()))

    @staticmethod
    def create_large_file(tarPath, compression, fileCount):
        # I have committed the resulting bz2 file to save test time.
        t0 = time.time()
        create_file = TestLibarchiveMountSource._create_file
        with tarfile.open(name=tarPath, mode='w:' + compression) as tarFile:
            for i in range(fileCount):
                create_file(tarFile, name=str(i), contents=str(i % 10))
                if i % 50_000 == 0:
                    print(f"Added {i} out of {fileCount} files to .tar.{compression} in {time.time() - t0:.3f} s")

        # contents = str(i)
        #   300k files for bz2 takes ~13 s and the resulting file is 986 KiB
        #   300k files for xz takes ~26 s and the resulting file is 259 KiB
        #   300k files for gz takes ~34 s and the resulting file is 2662 KiB
        # contents = str(i % 10)
        #   300k files for bz2 takes ~12 s and the resulting file is 580 KiB
        #   300k files for xz takes ~32 s and the resulting file is 544 KiB
        #   300k files for gz takes ~16 s and the resulting file is 2774 KiB
        # Funny how the compressed size is larger than the str(i) case for the LZ-based compressors
        # even though the uncompressed size is smaller! Only bz2 actually shrinks in compressed size!

    @staticmethod
    def _test_large_file(path):
        t0 = time.time()
        fileCount = 0
        with LibarchiveMountSource(path) as mountSource:
            t1 = time.time()
            print(f"Opening {path} took {time.time() - t0:.3f} s")  # ~5 s
            # In the worst case, reading all files can take 300k * 5s / 2 = ~9 days.
            # In the best case, it should take roughly 5s, same as when iterating over the archive in order.
            # The worst case happens if:
            #  - The returned order by list is messed up, e.g., random in the worst case, or even this is bad:
            #    0, 1, 10, 100, 1000, 10000, 100000, 100001, 100002, ...
            #  - Each file open reads the archive from the beginning instead of reusing the current libarchive handle.
            entries = mountSource.list('/')
            assert isinstance(entries, dict)
            t2 = time.time()
            print(f"Listing all {len(entries)} files took {t2 - t1:.3f} s")  # ~2 s

            fileCount = 0
            for fileName, fileInfo in entries.items():
                with mountSource.open(fileInfo) as file:
                    assert file.read() == fileName[-1:].encode()

                if fileCount > 0 and fileCount % 50_000 == 0:
                    print(f"Checked {fileCount} files' contents.")
                fileCount += 1

                # Depends on the system, but accounting for 10 times slower systems should be a good margin.
                # And we are still FAR off from the worst case time that should not happen.
                assert time.time() - t2 < 360

            print(f"Reading all {fileCount} files took {time.time() - t2:.3f} s")  # 36 s on my system

            # Test seeking back for good measure
            fileName = '2'
            fileInfo = entries.get(fileName)
            with mountSource.open(fileInfo) as file:
                assert file.read() == fileName[-1:].encode()

    @staticmethod
    @pytest.mark.parametrize('compression', ['bz2', 'gz', 'xz'])
    def test_large_file(compression):
        path = "tar-with-300-folders-with-1000-files-1B-files.tar." + compression
        with tempfile.NamedTemporaryFile(suffix='.' + path) as tmpTarFile:
            TestLibarchiveMountSource.create_large_file(tmpTarFile.name, compression=compression, fileCount=300_000)
            TestLibarchiveMountSource._test_large_file(tmpTarFile.name)



# pylint: disable=abstract-method

import shutil
import tempfile
from typing import IO, Any, Callable, Dict, Iterable, List, Optional, Union

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex
from ratarmountcore.utils import RatarmountError, overrides


class SQLiteIndexMountSource(MountSource):
    def __init__(
        self,
        index: Union[SQLiteIndex, str, IO[bytes]],
        clearIndexCache: bool = False,
        checkMetadata: Optional[Callable[[Dict[str, Any]], None]] = None,
        printDebug: int = 0,
        **_,
    ) -> None:
        self.indexFilePath = ""

        if isinstance(index, SQLiteIndex):
            self.index = index
            if clearIndexCache:
                self.index.clear_indexes()
            self.index.open_existing(checkMetadata=checkMetadata)
        else:
            # Open existing index without any corresponding archive, i.e., file open will not work!
            if isinstance(index, str):
                with open(index, 'rb') as file:
                    SQLiteIndexMountSource._quick_check_file(file, index)
                self.indexFilePath = index
            else:
                SQLiteIndexMountSource._quick_check_file(index, "File object")

                # Copy to a temporary file because sqlite cannot work with Python file objects. This can be wasteful!
                index.seek(0)
                self._temporaryFile = tempfile.NamedTemporaryFile(suffix=".ratarmount.index.sqlite", delete=True)
                shutil.copyfileobj(index, self._temporaryFile.file)  # type: ignore
                self._temporaryFile.file.flush()

                self.indexFilePath = self._temporaryFile.name

            # Encoding is only used for set_file_infos, so we are fine not forwarding it.
            self.index = SQLiteIndex(
                indexFilePath=self.indexFilePath, indexFolders=[], deleteInvalidIndexes=False, printDebug=printDebug
            )
            self.index.open_existing(checkMetadata=checkMetadata, readOnly=True)
            if not self.index.index_is_loaded():
                raise RatarmountError(f"Specified file {self.indexFilePath} is not a valid Ratarmount index.")

    @staticmethod
    def _quick_check_file(fileObject: IO[bytes], name: str) -> None:
        try:
            if fileObject.read(len(SQLiteIndex.MAGIC_BYTES)) == SQLiteIndex.MAGIC_BYTES:
                return
        finally:
            fileObject.seek(0)

        raise RatarmountError(name + " is not an ratarmount index file.")

    @staticmethod
    def _check_database(connection) -> bool:
        # May throw when sqlar does not exist or it is encrypted without the correct key being specified.
        result = connection.execute("SELECT name FROM sqlar LIMIT 1;").fetchone()
        return result and result[0]

    def __enter__(self):
        return self

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.index.close()

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        return self.index.lookup(path, fileVersion=fileVersion)

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        raise NotImplementedError

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        return self.index.list(path)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        return self.index.list_mode(path)

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        versions = self.index.versions(path)
        return len(versions) if isinstance(versions, dict) else 0

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        return self.index.list_xattr(fileInfo)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        return self.index.get_xattr(fileInfo, key)



#!/usr/bin/env python3

# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import shutil
import stat
import subprocess
import sys
import tempfile
import time
import zipfile
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.zip import ZipMountSource  # noqa: E402

try:
    import pyminizip
except ImportError:
    pyminizip = None  # type: ignore


def find_test_file(relativePathOrName):
    for i in range(3):
        path = os.path.sep.join([".."] * i + ["tests", relativePathOrName])
        if os.path.exists(path):
            return path
    return relativePathOrName


password = "foo"


def create_encrypted_test_file(path):
    old_working_directory = os.getcwd()

    files = {
        "random-8MiB": os.urandom(8 * 1024 * 1024),
        "digits-8MiB": b'01234567' * (1024 * 1024),
    }

    folder = None
    try:
        folder = tempfile.TemporaryDirectory()
        os.chdir(folder.name)
        for name, contents in files.items():
            Path(name).write_bytes(contents)

        if os.path.exists(path):
            os.remove(path)

        command = []
        if shutil.which("zip"):
            command = ["zip", "--encrypt", "--password", password, str(path), *files.keys()]
        elif shutil.which("7z") or shutil.which("7z.exe"):
            binary = "7z" if shutil.which("7z") else "7z.exe"
            command = [binary, "a", "-p" + password, str(path), *files.keys()]
        elif pyminizip:
            print("Create zip file with pyminizip.")
            pyminizip.compress_multiple(list(files.keys()), [], str(path), password, 3)

        if command:
            print("Create zip file with:", " ".join(command))
            subprocess.run(command, check=True)
    finally:
        os.chdir(old_working_directory)
        if folder is not None:
            shutil.rmtree(folder.name)
    return files


class TestZipMountSource:
    @staticmethod
    def test_simple_usage():
        with ZipMountSource(find_test_file('folder-symlink.zip')) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'

            # Links are not resolved by the mount source but by FUSE, i.e., descending into a link to a folder
            # will not work. This behavior may change in the future.
            for linkPath in ['/foo/jet']:
                assert mountSource.lookup(linkPath)
                assert mountSource.versions(linkPath) == 1
                assert not mountSource.list(linkPath)
                fileInfo = mountSource.lookup(linkPath)
                assert fileInfo.linkname == 'fighter'
                with mountSource.open(mountSource.lookup(linkPath)) as file:
                    # Contents of symlink is the symlink destination itself.
                    # This behavior is not consistent with other MountSources and therefore subject to change!
                    assert file.read() == b'fighter'

    @staticmethod
    def test_transform():
        with ZipMountSource(find_test_file('folder-symlink.zip'), transform=("(.)/(.)", r"\1_\2")) as mountSource:
            print(mountSource.list("/").keys())
            for folder in ['/', '/foo', '/foo_fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)
                assert mountSource.versions(folder) == 1

            for filePath in ['/foo_fighter_ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'


def benchmark_fast_zipfile_decryption():
    with tempfile.NamedTemporaryFile(suffix=".zip") as archive_path:
        files = create_encrypted_test_file(archive_path.name)
        with zipfile.ZipFile(archive_path.name) as archive:
            archive.setpassword(password.encode())
            assert set(archive.namelist()) == set(files.keys())

            t0 = time.time()
            for name, contents in files.items():
                with archive.open(name, "r") as file:
                    assert file.read() == contents
            t1 = time.time()
            duration = t1 - t0

            # On my local system: 6.2 s with Python, 0.1 s with this fix, with fix and pytest: 1.1 s !?
            # On Github Actions: 7.6 s with Python, 0.1 s with this fix
            # Thanks to the huge performance difference, this check should be sufficiently stable.
            print(f"Decryption took: {duration:.2e} s")
            assert duration < 1.0


def benchmark_fast_decryption():
    with tempfile.NamedTemporaryFile(suffix=".zip") as archive_path:
        files = create_encrypted_test_file(archive_path.name)
        with ZipMountSource(archive_path.name, passwords=[password.encode()]) as mountSource:
            t0 = time.time()
            for name, contents in files.items():
                with mountSource.open(mountSource.lookup(name)) as file:
                    assert file.read() == contents
            t1 = time.time()
            duration = t1 - t0

            # On my local system: 6.2 s with Python, 0.1 s with this fix
            # On Github Actions: 7.6 s with Python, 0.1 s with this fix
            # Thanks to the huge performance difference, this check should be sufficiently stable.
            print(f"Decryption took: {duration:.2e} s")
            assert duration < 1.0


# We need to run these tests without pytest because, for some reason,
# pytest slows the zip decryption fix down from 0.1 s to 1.1 s?!
if __name__ == '__main__':
    benchmark_fast_zipfile_decryption()
    benchmark_fast_decryption()



import errno
import io
import os
import stat
from typing import IO, Dict, Iterable, Optional, Union

from ratarmountcore.formats import FileFormatID, replace_format_check
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.utils import overrides

try:
    import pyfatfs
    from pyfatfs import PyFATException
    from pyfatfs.FatIO import FatIO
    from pyfatfs.PyFat import PyFat
except ImportError:
    pyfatfs = None  # type: ignore
    FatIO = None  # type: ignore
    PyFat = None  # type: ignore
    PyFATException = None  # type: ignore


def is_fat_image(fileObject) -> bool:
    if PyFat is None:
        return False

    offset = fileObject.tell()
    try:
        fs = PyFat()
        # TODO Avoid possibly slow full FAT parsing here. Only do some quick checks such as PyFatFS.PyFat.parse_header
        #      Calling __set_fp instead of set_fp avoids that but it is not part of the public interface per convention!
        fs._PyFat__set_fp(fileObject)
        fs.is_read_only = True
        try:
            fs.parse_header()
            return True
        except (pyfatfs.PyFATException, ValueError):
            return False
        finally:
            # Reset file object so that it does not get closed! Cannot be None because that is checked.
            fs._PyFat__fp = io.BytesIO()

    finally:
        fileObject.seek(offset)


replace_format_check(FileFormatID.FAT, is_fat_image)


class FATMountSource(MountSource):
    def __init__(self, fileOrPath: Union[str, IO[bytes]], **options) -> None:
        if PyFat is None:
            raise ImportError("Failed to find pyfatfs. Try: pip install pyfatfs")

        self.fileSystem = PyFat()
        if isinstance(fileOrPath, str):
            # TODO Probably good idea for performance on Lustre to open the file unbuffered.
            self.fileSystem.open(fileOrPath, read_only=True)
        else:
            self.fileSystem.set_fp(fileOrPath)
        self.options = options

    @staticmethod
    def _convert_fatdirectory_entry_to_file_info(entry, path) -> FileInfo:
        """
        entry: of type pyfatfs.FATDirectoryEntry.FATDirectoryEntry.
        """
        # FAT has no file permissions.
        mode = 0o777 | (stat.S_IFDIR if entry.is_directory() else stat.S_IFREG)

        # fmt: off
        return FileInfo(
            size     = entry.filesize,
            mtime    = entry.get_mtime().timestamp(),
            mode     = mode,
            linkname = "",  # FAT has no support for hard or symbolic links
            uid      = os.getuid(),
            gid      = os.getgid(),
            userdata = [path],
        )
        # fmt: on

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    @overrides(MountSource)
    def exists(self, path: str) -> bool:
        try:
            self.fileSystem.root_dir.get_entry(path)
        except PyFATException as exception:
            if exception.errno == errno.ENOENT:
                return False
            raise exception
        return True

    def _list(self, path: str) -> Optional[Iterable]:
        try:
            directories, files, _ = self.fileSystem.root_dir.get_entry(os.path.normpath(path)).get_entries()
        except PyFATException as exception:
            if exception.errno in [errno.ENOENT, errno.ENOTDIR]:
                return None
            raise exception
        return [str(entry) for entry in directories + files]

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        # TODO I think with the low-level API, we could also get the FileInfos
        return self._list(path)

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        try:
            entry = self.fileSystem.root_dir.get_entry(path)
        except PyFATException as exception:
            if exception.errno in [errno.ENOTDIR, errno.ENOENT]:
                return None
            raise exception
        return self._convert_fatdirectory_entry_to_file_info(entry, path)

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return 1

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        path = fileInfo.userdata[-1]
        assert isinstance(path, str)
        # TODO There is no option in FatIO to configure the buffering yet.
        return FatIO(self.fileSystem, path)

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.fileSystem.close()



import io
import sqlite3
from typing import IO, Optional, cast

from .StenciledFile import JoinedFile, LambdaReaderFile
from .utils import overrides


# CPython 3.11 added blobopen! https://docs.python.org/3/library/sqlite3.html#sqlite3.Blob
# Well, will have to wait for some years to fully replace this for compatibility reasons.
class SQLiteBlobFile(LambdaReaderFile):
    """Provides a memory-efficient file object interface to a single large blob inside an SQLite table"""

    def __init__(self, connection: sqlite3.Connection, expression: str, size: Optional[int] = None):
        """
        expression: Should yield a single blob when executed. And should also contain format string placeholders {}
                    before and after the column name.
        """

        self.connection = connection
        self.expression = expression

        super().__init__(
            (lambda offset, size: SQLiteBlobFile.read_blob_part(self.connection, expression, offset, size)),
            self.connection.execute(self.expression.format("LENGTH(", ")")).fetchone()[0] if size is None else size,
        )

    @staticmethod
    def read_blob_part(connection: sqlite3.Connection, expression: str, offset: int, size: int):
        # Note that SQLite offsets begin counting at 1 unlike any other programming language except maybe for Fortran
        return connection.execute(expression.format("substr(", ",?,?)"), (offset + 1, size)).fetchone()[0]


class SQLiteBlobsFile(JoinedFile):
    """Provides a memory-efficient file object interface to multiple large blobs inside an SQLite table"""

    def __init__(
        self, connection: sqlite3.Connection, table: str, column: str, buffer_size: int = io.DEFAULT_BUFFER_SIZE
    ):
        super().__init__(
            [
                cast(
                    IO,
                    SQLiteBlobFile(
                        connection, f"SELECT {{}}{column}{{}} FROM {table} WHERE ROWID == {rowid}", size=size
                    ),
                )
                for rowid, size in connection.execute(f"SELECT ROWID,LENGTH({column}) FROM {table} ORDER BY ROWID")
            ],
            buffer_size=buffer_size,
        )


class WriteSQLiteBlobs(io.RawIOBase):
    def __init__(self, connection: sqlite3.Connection, table: str, blob_size: int = io.DEFAULT_BUFFER_SIZE) -> None:
        io.RawIOBase.__init__(self)
        self.connection = connection
        self.table = table
        self.blob_size = blob_size
        self.blob = io.BytesIO()
        self._size = 0

    def _flush_blob(self):
        if self.blob.tell() > 0:
            self.connection.execute(f'INSERT INTO {self.table} VALUES (?)', (self.blob.getbuffer(),))
        self.blob = io.BytesIO()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self._flush_blob()

    @overrides(io.RawIOBase)
    def close(self) -> None:
        self._flush_blob()

    @overrides(io.RawIOBase)
    def fileno(self) -> int:
        # This is a virtual Python level file object and therefore does not have a valid OS file descriptor!
        raise io.UnsupportedOperation

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return False

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return False

    @overrides(io.RawIOBase)
    def writable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def write(self, buffer) -> int:
        freeBytesInBlob = self.blob_size - self.blob.tell()
        writtenCount = 0

        if len(buffer) < freeBytesInBlob:
            writtenCount += self.blob.write(buffer)
        else:
            writtenCount += self.blob.write(buffer[:freeBytesInBlob])
            self._flush_blob()
            writtenCount += self.blob.write(buffer[freeBytesInBlob:])

        self._size += writtenCount

        if writtenCount != len(buffer):
            raise RuntimeError("Failed to write all of the given data out!")

        return len(buffer)

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        # We are always at SEEK_END because no real seeking is implemented.
        if (whence in [io.SEEK_CUR, io.SEEK_END] and offset == 0) or (whence == io.SEEK_SET and offset == self.tell()):
            return self._size
        raise io.UnsupportedOperation

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        return self._size



import os
import stat
from typing import IO, Any, Dict, Iterable, List, Optional, Union

from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.utils import overrides


def max_up_count(path):
    if os.path.isabs(path):
        return 0
    result = 0
    upCount = 0
    for part in path.split(os.path.sep):
        if part == '..':
            upCount += 1
            result = max(result, upCount)
        elif part in ['.', '']:
            continue
        else:
            upCount -= 1
    return result


class FolderMountSource(MountSource):
    """
    This class manages one folder as mount source offering methods for listing folders, reading files, and others.
    """

    def __init__(self, path: Union[str, os.PathLike]) -> None:
        self.root = str(path)
        self._statfs = FolderMountSource._get_statfs_for_folder(self.root)

    def set_folder_descriptor(self, fd: int) -> None:
        """
        Make this mount source manage the special "." folder by changing to that directory.
        Because we change to that directory it may only be used for one mount source but it also works
        when that mount source is mounted on!
        """
        os.fchdir(fd)
        self.root = '.'
        self._statfs = FolderMountSource._get_statfs_for_folder(self.root)

    @staticmethod
    def _get_statfs_for_folder(path: str):
        result = os.statvfs(path)
        return {
            'f_bsize': result.f_bsize,
            'f_frsize': result.f_frsize,
            'f_blocks': result.f_blocks,
            'f_bfree': 0,
            'f_bavail': 0,
            'f_files': result.f_files,
            'f_ffree': 0,
            'f_favail': 0,
            'f_namemax': result.f_namemax,
        }

    def _realpath(self, path: str) -> str:
        """Path given relative to folder root. Leading '/' is acceptable"""
        return os.path.join(self.root, *path.strip('/').split('/'))

    @staticmethod
    def _stats_to_file_info(stats: os.stat_result, path: str, linkname: str):
        # fmt: off
        return FileInfo(
            size     = stats.st_size,
            mtime    = stats.st_mtime,
            mode     = stats.st_mode,
            linkname = linkname,
            uid      = stats.st_uid,
            gid      = stats.st_gid,
            userdata = [path],
        )
        # fmt: on

    @staticmethod
    def _dir_entry_to_file_info(dirEntry: os.DirEntry, path: str, realpath: str):
        try:
            linkname = os.readlink(realpath) if dirEntry.is_symlink() else ""
        except OSError:
            linkname = ""

        return FolderMountSource._stats_to_file_info(dirEntry.stat(follow_symlinks=False), linkname, path)

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return False

    @overrides(MountSource)
    def exists(self, path: str) -> bool:
        return os.path.lexists(self._realpath(path))

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        """All returned file infos contain a file path string at the back of FileInfo.userdata."""

        # This is a bit of problematic design, however, the versions count from 1 for the user.
        # And as -1 means the last version, 0 should also mean the first version ...
        # Basically, I did accidentally mix user-visible versions 1+ versions with API 0+ versions,
        # leading to this problematic clash of 0 and 1.
        if fileVersion not in [0, 1] or not self.exists(path):
            return None

        realpath = self._realpath(path)
        linkname = ""
        if os.path.islink(realpath):
            linkname = os.readlink(realpath)
            # Resolve relative links that point outside the source folder because they will become invalid
            # if they are mounted onto a different path. This relatively simply logic only works under the
            # assumption that "path" is normalized, i.e., it does not contain links in its path and no double
            # slashes and no '/./'. Calling os.path.normpath would remedy the latter but ONLY under the
            # assumption that there are no symbolic links in the path, else it might make things worse.
            if (
                not os.path.isabs(linkname)
                and max_up_count(linkname) > path.strip('/').count('/')
                and os.path.exists(realpath)
            ):
                realpath = os.path.realpath(realpath)
                return self._stats_to_file_info(os.stat(realpath), realpath, "")
        return self._stats_to_file_info(os.lstat(realpath), path.lstrip('/'), linkname)

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        realpath = self._realpath(path)
        if not os.path.isdir(realpath):
            return None

        return {
            os.fsdecode(dirEntry.name): FolderMountSource._dir_entry_to_file_info(dirEntry, path, realpath)
            for dirEntry in os.scandir(realpath)
        }

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        realpath = self._realpath(path)
        if not os.path.isdir(realpath):
            return None

        # https://docs.python.org/3/library/os.html#os.scandir
        # > All os.DirEntry methods may perform a system call, but is_dir() and is_file() usually
        # > only require a system call for symbolic links; os.DirEntry.stat() always requires a
        # > system call on Unix but only requires one for symbolic links on Windows.
        # Unfortunately, I am not sure whether it would be sufficient to build the file mode from these
        # two getters. For now, I'd say that all the esoteric stuff is simply not supported.
        def make_mode(dirEntry):
            mode = stat.S_IFDIR if dirEntry.is_dir(follow_symlinks=False) else stat.S_IFREG
            if dirEntry.is_symlink():
                mode = stat.S_IFLNK
            return mode

        return {os.fsdecode(dirEntry.name): make_mode(dirEntry) for dirEntry in os.scandir(realpath)}

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return 1 if self.exists(path) else 0

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        realpath = self.get_file_path(fileInfo)
        try:
            return open(realpath, 'rb', buffering=buffering)
        except Exception as e:
            raise ValueError(f"Specified path '{realpath}' is not a file that can be read!") from e

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        return self._statfs.copy()

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        return os.listxattr(self.get_file_path(fileInfo), follow_symlinks=False) if hasattr(os, 'listxattr') else []

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        return (
            os.getxattr(self.get_file_path(fileInfo), key, follow_symlinks=False) if hasattr(os, 'getxattr') else None
        )

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

    def get_file_path(self, fileInfo: FileInfo) -> str:
        path = fileInfo.userdata[-1]
        assert isinstance(path, str)
        # Path argument is only expected to be absolute for symbolic links pointing outside self.root.
        return path if path.startswith('/') else self._realpath(path)



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.compressions import (  # noqa: E402
    check_for_sequence,
    has_matching_alphabets,
    strip_suffix_from_archive,
    strip_suffix_from_compressed_file,
)
from ratarmountcore.utils import ALPHA, DIGITS, HEX, format_number  # noqa: E402


def test_strip_suffix_from_compressed_file():
    ssc = strip_suffix_from_compressed_file

    assert ssc('a.tar.bz2') == 'a.tar'
    assert ssc('a.tar.BZ2') == 'a.tar'
    assert ssc('a.tar.BZIP2') == 'a.tar'
    assert ssc('a.tar.gz') == 'a.tar'
    assert ssc('a.tar.gzip') == 'a.tar'
    assert ssc('a.tar.xz') == 'a.tar'
    assert ssc('a.tar.zst') == 'a.tar'
    assert ssc('a.tar') == 'a.tar'
    assert ssc('a.mp3') == 'a.mp3'


def test_strip_suffix_from_archive():
    sst = strip_suffix_from_archive

    assert sst('a.tar.bz2') == 'a'
    assert sst('a.tar.BZ2') == 'a'
    assert sst('a.tar.BZIP2') == 'a'
    assert sst('a.tar.gz') == 'a'
    assert sst('a.tar.gzip') == 'a'
    assert sst('a.tar.xz') == 'a'
    assert sst('a.tar.zst') == 'a'
    assert sst('a.tar') == 'a'
    assert sst('a.mp3') == 'a.mp3'

    assert sst('a.tbz2') == 'a'
    assert sst('a.TBZ2') == 'a'
    assert sst('a.tgz') == 'a'
    assert sst('a.txz') == 'a'
    assert sst('a.tzst') == 'a'


def test_has_matching_alphabets():
    matches = has_matching_alphabets

    assert matches('a', 'b')
    assert matches('0', '1')
    assert matches('0', 'a')  # because both might be hexadecimal
    assert matches('1a', 'b0')
    assert matches(HEX, HEX)
    assert matches(DIGITS, DIGITS)
    assert matches(ALPHA, ALPHA)
    assert matches(HEX, DIGITS)
    assert not matches(ALPHA, HEX)
    assert not matches(ALPHA, DIGITS)
    assert not matches('ag', 'b0')


def test_check_for_sequence():
    def to_alpha1(i):
        return format_number(i, ALPHA, 1)

    assert check_for_sequence(['a'], to_alpha1) == ['a']
    assert check_for_sequence(['a', 'b'], to_alpha1) == ['a', 'b']
    assert check_for_sequence(['b', 'a'], to_alpha1) == ['a', 'b']
    assert check_for_sequence(['b', 'a', 'd'], to_alpha1) == ['a', 'b']
    assert check_for_sequence(['a', 'd'], to_alpha1) == ['a']
    assert check_for_sequence(['a', 'd', 'e'], to_alpha1) == ['a']
    assert not check_for_sequence(['aa'], to_alpha1)
    assert not check_for_sequence(['0'], to_alpha1)

    def to_alpha2(i):
        return format_number(i, ALPHA, 2)

    assert check_for_sequence(['aa'], to_alpha2) == ['aa']
    assert check_for_sequence(['aa', 'ab'], to_alpha2) == ['aa', 'ab']
    assert check_for_sequence(['ab', 'aa'], to_alpha2) == ['aa', 'ab']
    assert check_for_sequence(['ab', 'aa', 'ad'], to_alpha2) == ['aa', 'ab']
    assert check_for_sequence(['aa', 'ad'], to_alpha2) == ['aa']
    assert check_for_sequence(['aa', 'ad', 'ae'], to_alpha2) == ['aa']
    assert not check_for_sequence(['aaa'], to_alpha2)
    assert not check_for_sequence(['0'], to_alpha2)
    assert not check_for_sequence(['00'], to_alpha2)

    def to_digit3(i):
        return format_number(i, DIGITS, 3)

    assert check_for_sequence(['000'], to_digit3) == ['000']
    assert check_for_sequence(['000', '001'], to_digit3) == ['000', '001']
    assert check_for_sequence(['001', '000'], to_digit3) == ['000', '001']
    assert check_for_sequence(['001', '000', '003'], to_digit3) == ['000', '001']
    assert check_for_sequence(['000', '003'], to_digit3) == ['000']
    assert check_for_sequence(['000', '003', '004'], to_digit3) == ['000']
    assert not check_for_sequence(['0001'], to_digit3)



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import stat
import sys

from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource  # noqa: E402


class TestSQLiteIndexMountSource:
    @staticmethod
    def test_password():
        with copy_test_file("nested-tar.index.sqlite") as path, SQLiteIndexMountSource(path) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)
                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo', '/foo/lighter.tar']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)

            # File contents cannot be read from a mounted index alone!

            # The 'transform' parameter is ignored because it is only for index creation, not index loading!



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import importlib
import io
import os
import stat
import struct
import sys

import pytest
from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.formats import find_squashfs_offset  # noqa: E402
from ratarmountcore.mountsource.formats.squashfs import SquashFSMountSource  # noqa: E402

compressionsToTest = []
if importlib.util.find_spec('PySquashfsImage'):
    compressionsToTest = ['no-compression', 'gzip', 'lzma', 'xz']
    if importlib.util.find_spec('lz4') is not None:
        compressionsToTest.append('lz4')
    if importlib.util.find_spec('lzo') is not None:
        compressionsToTest.append('lzo')
    if importlib.util.find_spec('zstandard') is not None:
        compressionsToTest.append('zstd')


class TestSquashfsMountSource:
    @staticmethod
    def test_find_magic_bytes():
        assert find_squashfs_offset(io.BytesIO()) < 0
        assert find_squashfs_offset(io.BytesIO(b"")) < 0
        assert find_squashfs_offset(io.BytesIO(b"a")) < 0
        assert find_squashfs_offset(io.BytesIO(b"ab")) < 0
        assert find_squashfs_offset(io.BytesIO(b"ab")) < 0
        assert find_squashfs_offset(io.BytesIO(b"foob")) < 0

        validHeader = b"hsqs" + struct.pack('<IIII', 0, 0, 4096, 0) + struct.pack('<HHHHHH', 0, 12, 0, 0, 4, 0)
        assert find_squashfs_offset(io.BytesIO(validHeader)) == 0
        assert find_squashfs_offset(io.BytesIO(b"0" + validHeader)) == 1
        assert find_squashfs_offset(io.BytesIO(b"0" * 1234 + validHeader)) == 1234
        assert find_squashfs_offset(io.BytesIO(b"0" * 1234 + validHeader + validHeader)) == 1234

    @staticmethod
    @pytest.mark.parametrize('compression', compressionsToTest)
    def test_simple_usage(compression):
        with copy_test_file(f'folder-symlink.{compression}.squashfs') as path, SquashFSMountSource(path) as mountSource:
            with open(path, 'rb') as file:
                assert find_squashfs_offset(file) == 0

            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)

                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read(1) == b'i'
                    assert file.read(5) == b'riya\n'

                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'
                    assert file.seek(0) == 0
                    assert file.read() == b'iriya\n'
                    assert file.seek(1) == 1
                    assert file.read() == b'riya\n'
                    assert file.seek(5) == 5
                    assert file.read() == b'\n'
                    assert file.seek(6) == 6
                    assert file.tell() == 6
                    assert file.read() == b''
                    assert file.tell() == 6
                    assert file.read(2) == b''
                    assert file.tell() == 6

            # Links are not resolved by the mount source but by FUSE, i.e., descending into a link to a folder
            # will not work. This behavior may change in the future.
            for linkPath in ['/foo/jet']:
                assert mountSource.lookup(linkPath)
                assert mountSource.versions(linkPath) == 1
                assert not mountSource.list(linkPath)
                fileInfo = mountSource.lookup(linkPath)
                assert fileInfo.linkname == 'fighter'

    @staticmethod
    @pytest.mark.parametrize('compression', compressionsToTest)
    def test_transform(compression):
        with copy_test_file(f'folder-symlink.{compression}.squashfs') as path, SquashFSMountSource(
            path, transform=("(.)/(.)", r"\1_\2")
        ) as mountSource:
            for folder in ['/', '/foo', '/foo_fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)
                assert mountSource.versions(folder) == 1

            for filePath in ['/foo_fighter_ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)



import os
from typing import IO, Dict, Iterable, List, Optional, Tuple, Union

from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info
from ratarmountcore.utils import overrides


class SubvolumesMountSource(MountSource):
    def __init__(self, mountSources: Dict[str, MountSource], printDebug: int = 0) -> None:
        """
        mountSources : List of mount sources to mount as subfolders.
        """
        self.mountSources: Dict[str, MountSource] = mountSources
        self.printDebug = printDebug

        for name in self.mountSources:
            if not name:
                raise ValueError("Mount source names may not be empty!")
            if '/' in name:
                raise ValueError(f"Mount source names may not contain slashes! ({name})")

        self.rootFileInfo = create_root_file_info(userdata=[None])

    def _find_mount_source(self, path: str) -> Optional[Tuple[str, str]]:
        path = path.strip('/')
        subvolume, subpath = path.split('/', maxsplit=1) if '/' in path else [path, ""]
        return (subvolume, '/' + subpath) if subvolume in self.mountSources else None

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return all(m.is_immutable() for m in self.mountSources.values())

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        path = path.strip('/')
        if not path:
            return self.rootFileInfo.clone()

        result = self._find_mount_source(path)
        if result is None:
            return None
        subvolume, subpath = result

        fileInfo = self.mountSources[subvolume].lookup(subpath, fileVersion=fileVersion)
        if isinstance(fileInfo, FileInfo):
            fileInfo.userdata.append(subvolume)
            return fileInfo

        return None

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        path = path.strip('/')
        if not path:
            return 1

        result = self._find_mount_source(path)
        if result is None:
            return 0
        subvolume, subpath = result

        return self.mountSources[subvolume].versions(subpath)

    def _list(self, path: str, onlyMode: bool):
        path = path.strip('/')
        if not path:
            return dict.fromkeys(self.mountSources.keys(), self.rootFileInfo.mode if onlyMode else self.rootFileInfo)

        result = self._find_mount_source(path)
        if result is None:
            return None
        subvolume, subpath = result

        return (
            self.mountSources[subvolume].list_mode(subpath) if onlyMode else self.mountSources[subvolume].list(subpath)
        )

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        return self._list(path, onlyMode=False)

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        return self._list(path, onlyMode=True)

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        subvolume = fileInfo.userdata.pop()
        if subvolume is None:
            raise ValueError(f"Found subvolume is None for fileInfo: {fileInfo}")
        try:
            return self.mountSources[subvolume].open(fileInfo, buffering=buffering)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        subvolume = fileInfo.userdata.pop()
        if subvolume is None:
            raise ValueError(f"Found subvolume is None for fileInfo: {fileInfo}")
        try:
            return self.mountSources[subvolume].read(fileInfo, size, offset)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        subvolume = fileInfo.userdata.pop()
        try:
            return [] if subvolume is None else self.mountSources[subvolume].list_xattr(fileInfo)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        subvolume = fileInfo.userdata.pop()
        try:
            return None if subvolume is None else self.mountSources[subvolume].get_xattr(fileInfo, key)
        finally:
            fileInfo.userdata.append(subvolume)

    @overrides(MountSource)
    def get_mount_source(self, fileInfo: FileInfo) -> Tuple[str, MountSource, FileInfo]:
        sourceFileInfo = fileInfo.clone()
        subvolume = sourceFileInfo.userdata.pop()

        if subvolume is None or subvolume not in self.mountSources:
            return '/', self, fileInfo
        mountSource = self.mountSources[subvolume]

        subpath, subMountSource, subFileInfo = mountSource.get_mount_source(sourceFileInfo)
        return os.path.normpath(f"/{subvolume}/{subpath}"), subMountSource, subFileInfo

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        for mountSource in self.mountSources.values():
            mountSource.__exit__(exception_type, exception_value, exception_traceback)

    def join_threads(self):
        for mountSource in self.mountSources:
            if hasattr(mountSource, 'join_threads'):
                mountSource.join_threads()



import dataclasses
import traceback
from typing import IO, Any, Callable, Dict, List, Optional, Set, Tuple, Union

from ratarmountcore.formats import FileFormatID

from . import MountSource
from .formats.asar import ASARMountSource
from .formats.ext4 import EXT4MountSource
from .formats.fat import FATMountSource
from .formats.libarchive import LibarchiveMountSource
from .formats.py7zr import Py7zrMountSource
from .formats.rar import RarMountSource
from .formats.sqlar import SQLARMountSource
from .formats.squashfs import SquashFSMountSource
from .formats.tar import SQLiteIndexedTar
from .formats.zip import ZipMountSource
from .SQLiteIndexMountSource import SQLiteIndexMountSource

try:
    import libarchive
except (ImportError, AttributeError):
    libarchive = None  # type: ignore

FID = FileFormatID


def _open_tar_mount_source(fileOrPath: Union[str, IO[bytes]], **options) -> MountSource:
    if isinstance(fileOrPath, str):
        if 'tarFileName' in options:
            copiedOptions = options.copy()
            del copiedOptions['tarFileName']
            return SQLiteIndexedTar(fileOrPath, **copiedOptions)
        return SQLiteIndexedTar(fileOrPath, **options)
    return SQLiteIndexedTar(fileObject=fileOrPath, **options)


def _open_libarchive_mount_source(fileOrPath: Union[str, IO[bytes]], **options) -> Optional[MountSource]:
    if libarchive is None:
        return None

    printDebug = int(options.get("printDebug", 0)) if isinstance(options.get("printDebug", 0), int) else 0

    try:
        if printDebug >= 2:
            print("[Info] Trying to open archive with libarchive backend.")
        return LibarchiveMountSource(fileOrPath, **options)
    except Exception as exception:
        if printDebug >= 2:
            print("[Info] Checking for libarchive file raised an exception:", exception)
        if printDebug >= 3:
            traceback.print_exc()
    finally:
        try:
            if hasattr(fileOrPath, 'seek'):
                fileOrPath.seek(0)  # type: ignore
        except Exception as exception:
            if printDebug >= 1:
                print("[Info] seek(0) raised an exception:", exception)
            if printDebug >= 2:
                traceback.print_exc()
    return None


@dataclasses.dataclass
class ArchiveBackendInfo:
    # Opens a MountSource from a path or file object and additional options (kwargs).
    # Note that the MountSource derived classes themselves (or rather their __init__) are fitting callables!
    open: Callable[[Any], Optional[MountSource]]
    # Supported file formats. These are for quick checks and prioritization based on file extension.
    formats: Set[FileFormatID]
    # If a format is suspected e.g. by extension or by a non-module dependent format check,
    # the modules listed here are checked and the module package name can be suggested to be installed.
    # Tuple: (module name, package name)
    requiredModules: List[Tuple[str, str]]


# Map of backends to their respective open-function. The order implies a priority.
# The priority is overwritten by the associated file formats for a backend, which is used to look up
# file extensions via ARCHIVE_FORMATS. Furthermore, the backend is only tried if any of the associated
# file format checkers returns True!
# The keys are the backend names the user can specify with --backends or via prioritizedBackends arguments.
ARCHIVE_BACKENDS: Dict[str, ArchiveBackendInfo] = {
    "rarfile": ArchiveBackendInfo(RarMountSource, {FID.RAR}, [('rarfile', 'rarfile')]),
    "tarfile": ArchiveBackendInfo(
        _open_tar_mount_source, {FID.TAR, FID.GZIP, FID.BZIP2, FID.XZ, FID.ZSTANDARD}, [('tarfile', '')]
    ),
    "zipfile": ArchiveBackendInfo(ZipMountSource, {FID.ZIP}, [('zipfile', '')]),
    "PySquashfsImage": ArchiveBackendInfo(
        SquashFSMountSource, {FID.SQUASHFS}, [('PySquashfsImage', 'PySquashfsImage')]
    ),
    # Formats are handily listed either in libarchive:
    # https://github.com/libarchive/libarchive/blob/6110e9c82d8ba830c3440f36b990483ceaaea52c/libarchive/
    #   archive_read_support_format_all.c#L32
    # Or in python-libarchive-c:
    # https://github.com/Changaco/python-libarchive-c/blob/5f7008d876103bac84c40905d00bb6b5afbab91a/libarchive/
    #   ffi.py#L243
    # Supported by other backends: tar, rar, zip
    # Not supported because it has no magic identification and therefore is only trouble: lha, mtree
    # http://fileformats.archiveteam.org/wiki/LHA
    # > LHA can be identified with high accuracy, but doing so can be laborious,
    # > due to the lack of a signature, and other complicating factors.
    "libarchive": ArchiveBackendInfo(
        _open_libarchive_mount_source,
        {
            FID.AR,
            FID.CAB,
            FID.XAR,
            FID.CPIO,
            FID.ISO9660,
            FID.WARC,
            # py7zr cannot handle symbolic links, therefore prefer libarchive!
            # But, libarchive does not support decrpytion.
            FID.SEVEN_ZIP,
            # Archive formats supported by other backends with higher precedence
            FID.RAR,
            FID.TAR,
            FID.ZIP,
            # Compression formats
            FID.BZIP2,
            FID.GZIP,
            FID.XZ,
            FID.ZSTANDARD,
            FID.GRZIP,
            FID.LRZIP,
            FID.LZ4,
            FID.LZIP,
            FID.LZMA,
            FID.LZOP,
            FID.RPM,
            FID.UU,
            FID.Z,
        },
        [('libarchive', 'libarchive-c')],
    ),
    "py7zr": ArchiveBackendInfo(Py7zrMountSource, {FID.SEVEN_ZIP}, [('py7zr', 'py7zr')]),
    "pyfatfs": ArchiveBackendInfo(FATMountSource, {FID.FAT}, [('pyfatfs', 'pyfatfs')]),
    "ext4": ArchiveBackendInfo(EXT4MountSource, {FID.EXT4}, [('ext4', 'ext4')]),
    "asar": ArchiveBackendInfo(ASARMountSource, {FID.ASAR}, []),
    "sqlar": ArchiveBackendInfo(SQLARMountSource, {FID.SQLAR}, []),
    "RatarmountIndex": ArchiveBackendInfo(SQLiteIndexMountSource, {FID.RATARMOUNT_INDEX}, []),
}



import contextlib
import datetime
import os
import stat
import sys
import time
from typing import IO, Dict, Iterable, List, Optional, Union

from ratarmountcore.formats import FileFormatID, replace_format_check
from ratarmountcore.mountsource import FileInfo, MountSource, create_root_file_info
from ratarmountcore.utils import overrides

with contextlib.suppress(ImportError):
    import rarfile


def is_rar_file(fileObject: IO[bytes]) -> bool:
    # @see https://www.rarlab.com/technote.htm#rarsign
    # > RAR 5.0 signature consists of 8 bytes: 0x52 0x61 0x72 0x21 0x1A 0x07 0x01 0x00.
    # > You need to search for this signature in supposed archive from beginning and up to maximum SFX module size.
    # > Just for comparison this is RAR 4.x 7 byte length signature: 0x52 0x61 0x72 0x21 0x1A 0x07 0x00.
    # > Self-extracting module (SFX)
    # > Any data preceding the archive signature. Self-extracting module size and contents is not defined.
    # > At the moment of writing this documentation RAR assumes the maximum SFX module size to not exceed 1 MB,
    # > but this value can be increased in the future.
    oldPosition = fileObject.tell()
    if fileObject.read(6) == b'Rar!\x1a\x07':
        return True
    if 'rarfile' in sys.modules:
        fileObject.seek(oldPosition)
        fileObject.seek(oldPosition)
        if rarfile.is_rarfile_sfx(fileObject):
            return True
    return False


replace_format_check(FileFormatID.RAR, is_rar_file)


class RarMountSource(MountSource):
    # Basically copy paste of ZipMountSource because the interfaces are very similar.
    def __init__(self, fileOrPath: Union[str, IO[bytes]], **options) -> None:
        if 'rarfile' not in sys.modules:
            raise RuntimeError("Did not find the rarfile module. Try: pip install rarfile")

        self.fileObject = rarfile.RarFile(fileOrPath, 'r')
        RarMountSource._find_password(self.fileObject, options.get("passwords", []))

        self.files = {RarMountSource._clean_path(info.filename): info for info in self.fileObject.infolist()}
        self.options = options

    @staticmethod
    def _clean_path(path):
        result = os.path.normpath(path) + ('/' if path.endswith('/') else '')
        while result.startswith('../'):
            result = result[3:]
        return result

    @staticmethod
    def _find_password(fileobj: "rarfile.RarFile", passwords):
        if not fileobj.needs_password():
            return None

        # If headers are encrypted, then infolist will simply return an empty list!
        files = fileobj.infolist()
        if not files:
            for password in passwords:
                fileobj.setpassword(password)
                files = fileobj.infolist()
                if files:
                    return password

        # If headers are not encrypted, then try out passwords by trying to open the first file.
        files = [file for file in files if file.is_file()]
        if not files:
            return None
        for password in passwords:
            fileobj.setpassword(password)
            try:
                with fileobj.open(files[0]) as file:
                    file.read(1)
                return password
            except (rarfile.PasswordRequired, rarfile.BadRarFile):
                pass

        raise rarfile.PasswordRequired("Could not find a matching password!")

    @staticmethod
    def _get_mode(info: "rarfile.RarInfo") -> int:
        return info.mode | (stat.S_IFLNK if info.file_redir else (stat.S_IFDIR if info.is_dir() else stat.S_IFREG))

    @staticmethod
    def _convert_to_file_info(normalizedPath: str, info: "rarfile.RarInfo") -> FileInfo:
        if info.date_time:
            dtime = datetime.datetime(*info.date_time)
            dtime = dtime.replace(tzinfo=datetime.timezone.utc)
            mtime = dtime.timestamp() if info.date_time else 0
        else:
            mtime = 0

        # fmt: off
        return FileInfo(
            size     = info.file_size,
            mtime    = mtime,
            mode     = RarMountSource._get_mode(info),
            # file_redir is (type, flags, target) or None. Only tested for type == RAR5_XREDIR_UNIX_SYMLINK.
            linkname = info.file_redir[2] if info.file_redir else "",
            uid      = os.getuid(),
            gid      = os.getgid(),
            userdata = [(normalizedPath, info)],
        )
        # fmt: on

    @overrides(MountSource)
    def is_immutable(self) -> bool:
        return True

    # TODO How to behave with files in archive with absolute paths? Currently, they would never be shown.
    @staticmethod
    def _get_name(folderPath, filePath):
        if not filePath.startswith(folderPath):
            return None

        filePath = filePath[len(folderPath) :].strip('/')
        if not filePath:
            return None

        # This effectively adds all parent paths as folders. It is easy to create
        # RARs and ZIPs with nested files without information on the parent directories!
        if '/' in filePath:
            firstSlash = filePath.index('/')
            filePath = filePath[:firstSlash]

        return filePath

    @overrides(MountSource)
    def list(self, path: str) -> Optional[Union[Iterable[str], Dict[str, FileInfo]]]:
        path = path.strip('/')
        if path:
            path += '/'

        # The "filename" member is wrongly named as it returns the full path inside the archive not just the name part.
        return {
            self._get_name(path, normalizedPath): self._convert_to_file_info(normalizedPath, info)
            for normalizedPath, info in self.files.items()
            if self._get_name(path, normalizedPath)
        }

    @overrides(MountSource)
    def list_mode(self, path: str) -> Optional[Union[Iterable[str], Dict[str, int]]]:
        path = path.strip('/')
        if path:
            path += '/'

        # The "filename" member is wrongly named as it returns the full path inside the archive not just the name part.
        return {
            self._get_name(path, normalizedPath): RarMountSource._get_mode(info)
            for normalizedPath, info in self.files.items()
            if self._get_name(path, normalizedPath)
        }

    def _lookups(self, path: str) -> List[FileInfo]:
        # If we have a fileInfo for the given directory path, then everything is fine.
        pathAsDir = path.strip('/') + '/'
        if pathAsDir == '/':
            return [create_root_file_info(userdata=[None])]

        infoList = [
            RarMountSource._convert_to_file_info(normalizedPath, info)
            for normalizedPath, info in self.files.items()
            if normalizedPath.rstrip('/') == path.lstrip('/')
        ]

        # Check whether some parent directories of files do not exist as separate entities in the archive.
        if not any(info.userdata[-1][0] == pathAsDir for info in infoList) and any(
            normalizedPath.rstrip('/').startswith(pathAsDir) for normalizedPath, info in self.files.items()
        ):
            # fmt: off
            infoList.append(
                FileInfo(
                    size     = 0,
                    mtime    = int(time.time()),
                    mode     = 0o777 | stat.S_IFDIR,
                    linkname = "",
                    uid      = os.getuid(),
                    gid      = os.getgid(),
                    userdata = [None],
                )
            )
            # fmt: on

        return infoList

    @overrides(MountSource)
    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        infos = self._lookups(path)
        return infos[fileVersion] if -len(infos) <= fileVersion < len(infos) else None

    @overrides(MountSource)
    def versions(self, path: str) -> int:
        return len(self._lookups(path))

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        # I do not see any obvious option to rarfile.RarFile to apply the specified buffer size.
        info = fileInfo.userdata[-1][1]
        assert isinstance(info, rarfile.RarInfo)
        return self.fileObject.open(info, 'r')

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.fileObject.close()



import time


class ProgressBar:
    """Simple progress bar which keeps track of changes and prints the progress and a time estimate."""

    def __init__(self, maxValue: float):
        # Do not use thread_time here even if it is twice as fast because it does not count the time the thread
        # has been sleeping, which it does when it waits for I/O. This made the times incorrect for slow HDD accesses!
        self._get_time = time.time
        self.maxValue = maxValue
        self.lastUpdateTime = self._get_time()
        self.lastUpdateValue = 0.0
        self.updateInterval = 2.0  # seconds
        self.creationTime = self._get_time()

    def update(self, value: float) -> None:
        """Should be called whenever the monitored value changes. The progress bar is updated accordingly."""
        if self.lastUpdateTime is not None and (self._get_time() - self.lastUpdateTime) < self.updateInterval:
            return

        # Use whole interval since start to estimate time
        percent = value / self.maxValue if self.maxValue != 0 else 1.0
        totalTime = self._get_time() - self.creationTime
        eta1 = int(totalTime / percent - totalTime if percent != 0 else 0)
        # Use only a shorter window interval to estimate time.
        # Accounts better for higher speeds in beginning, e.g., caused by caching effects.
        # However, this estimate might vary a lot while the other one stabilizes after some time!
        if value == self.lastUpdateValue:
            eta2 = 99999 * 60
        else:
            eta2 = int(
                (self._get_time() - self.lastUpdateTime) / (value - self.lastUpdateValue) * (self.maxValue - value)
            )
        print(
            f"Position {value} of {self.maxValue} ({percent * 100.0:.2f}%). "
            f"Remaining time: {eta2 // 60} min {eta2 % 60} s (current rate), "
            f"{eta1 // 60} min {eta1 % 60} s (average rate). "
            f"Spent time: {int(totalTime) // 60} min {int(totalTime) % 60} s",
            flush=True,
        )

        self.lastUpdateTime = self._get_time()
        self.lastUpdateValue = value



# libarchive.ffi produces many false positives on my local system and there seems to be nothing else works.
# pylint: disable=no-member

import contextlib
import ctypes
import io
import json
import os
import re
import stat
import sys
import tarfile
from timeit import default_timer as timer
from typing import IO, Any, Callable, Dict, List, Optional, Tuple, Union, cast

from ratarmountcore.compressions import COMPRESSION_BACKENDS
from ratarmountcore.formats import FILE_FORMATS
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.utils import InvalidIndexError, overrides

try:
    # Use the FFI directly because the higher-level interface does not work sufficiently for our use case.
    import libarchive.ffi as laffi
    from libarchive.exception import ArchiveError
except (ImportError, AttributeError):
    pass

try:
    import py7zr  # pylint: disable=unused-import
except ImportError:
    py7zr = None  # type: ignore


class ArchiveEntry:
    def __init__(self, archive, entry_index: int, encoding: str = 'utf-8'):
        self._archive = archive  # Store for lifetime
        self._entry = laffi.entry_new()
        self.eof = laffi.read_next_header2(self._archive, self._entry) == laffi.ARCHIVE_EOF
        self.encoding = encoding
        self.entry_index = entry_index
        self._fileInfoRow: Optional[Tuple] = None

    def __del__(self):
        laffi.entry_free(self._entry)

    def format_name(self):
        return laffi.format_name(self._archive)

    def get_time_by_name(self, name):
        if not getattr(laffi, f'entry_{name}_is_set')(self._entry):
            return None
        seconds = getattr(laffi, f'entry_{name}')(self._entry)
        nseconds = getattr(laffi, f'entry_{name}_nsec')(self._entry)
        return float(seconds) + float(nseconds) / 1e9 if nseconds else int(seconds)

    def get_time(self):
        result = self.get_time_by_name('mtime')
        if result is None:
            result = self.get_time_by_name('ctime')
        if result is None:
            result = self.get_time_by_name('birthtime')
        return result or 0

    def path(self) -> str:
        path = laffi.entry_pathname_w(self._entry)
        if not path:
            path = laffi.entry_pathname(self._entry)
            if path is not None:
                with contextlib.suppress(UnicodeError):
                    path = path.decode(self.encoding)
        return path

    def mode(self):
        return laffi.entry_mode(self._entry)

    def filetype(self):
        return laffi.entry_filetype(self._entry)

    def is_directory(self):
        return self.filetype() & 0o170000 == 0o040000

    def is_symbolic_link(self):
        return self.filetype() & 0o170000 == 0o120000

    def linkname(self):
        path = ""
        if self.is_symbolic_link():
            path = laffi.entry_symlink_w(self._entry)
            if not path:
                path = laffi.entry_symlink(self._entry)
        else:
            path = laffi.entry_hardlink_w(self._entry)
            if not path:
                path = laffi.entry_hardlink(self._entry)

        if isinstance(path, bytes):
            with contextlib.suppress(UnicodeError):
                path = path.decode(self.encoding)

        return path

    def convert_to_row(self, entryCount: int, transform: Callable[[str], str], path: Optional[str] = None) -> Tuple:
        # The data logic may only be evaluated once because determining the size may require reading the whole file!
        if self._fileInfoRow is not None:
            return self._fileInfoRow

        if laffi.entry_size_is_set(self._entry):
            size = laffi.entry_size(self._entry)
        else:
            buffer = ctypes.create_string_buffer(1024 * 1024)
            size = 0
            while True:
                readSize = laffi.read_data(self._archive, buffer, len(buffer))
                if not readSize:
                    break
                size += readSize

        mode = self.mode() & 0o777
        linkname = self.linkname()
        if linkname:
            mode = mode | stat.S_IFLNK
        else:
            linkname = ""
            mode = mode | (stat.S_IFDIR if self.is_directory() else stat.S_IFREG)

        if not path:
            path = self.path()
        path, name = SQLiteIndex.normpath(transform(path)).rsplit("/", 1)

        # Currently, this is unused. Only the headerOffset should be used.
        dataOffset = 0

        # The header offset needs to be set to some unique value in order to account for entries with identical paths.
        # We probably are not able to query the actual offset in the TAR via libarchive.
        # This makes the created index incompatible with the SQLiteIndexedTar backend.
        # For that reason more consistency checks were added as well as the backendName key in the index metadata.
        headerOffset = entryCount

        # fmt: off
        self._fileInfoRow = (
            path          ,  # 0  : path
            name          ,  # 1  : file name
            headerOffset  ,  # 2  : header offset
            dataOffset    ,  # 3  : data offset
            size          ,  # 4  : file size
            self.get_time(),  # 5  : modification time
            mode          ,  # 6  : file mode / permissions
            0             ,  # 7  : TAR file type. Currently unused. Overlaps with mode
            linkname      ,  # 8  : linkname
            0             ,  # 9  : user ID
            0             ,  # 10 : group ID
            False         ,  # 11 : is TAR (unused?)
            False         ,  # 12 : is sparse
            False         ,  # 13 : is generated (parent folder)
            0             ,  # 14 : recursion depth
        )
        # fmt: on

        return self._fileInfoRow


class IterableArchive:
    """A class that opens an archive with libarchive and can iterate over the entries."""

    # Read formats list supported by python-libarchive-c:
    #   '7zip', 'all', 'ar', 'cab', 'cpio', 'empty', 'iso9660', 'lha', 'mtree',
    #   'rar', 'raw', 'tar', 'xar', 'zip', 'warc'
    # We especially do not want mtree because it has no magic bytes and basically matches anything with newlines,
    # including some test text files. 'raw', 'empty', and 'all' are special input formats.
    # It seems that 'raw' cannot be combined in an or-manner. It disables all archive format recognitions and only
    # applies the filters.
    # https://github.com/libarchive/libarchive/wiki/FormatRaw#caveat-dont-mix-_raw_-with-other-handlers
    # > aveat: Don't mix _raw_ with other handlers
    # > If you are using the raw handler, you should generally not enable any other handler
    # Note that the lz4 and zstd filters are not available in the manylinux2014 container (CentOS Linux
    # release 7.9.2009 (Core))
    ENABLED_FORMATS = ('7zip', 'ar', 'cab', 'cpio', 'iso9660', 'lha', 'rar', 'rar5', 'tar', 'xar', 'warc', 'zip')

    def __init__(
        self,
        file: Union[str, IO[bytes]],
        encoding='utf-8',
        passwords: Optional[List[Union[str, bytes]]] = None,
        bufferSize=1024 * 1024,
        printDebug: int = 0,
    ):
        self.encoding = encoding
        self.printDebug = printDebug
        self.bufferSize = bufferSize
        self._file = file
        self._entryIndex = 0  # A consecutive number to identify entries even if they have the same name.
        self._buffer = None
        self._readCallback = None
        self._seekCallback = None
        self._entry: Optional[ArchiveEntry] = None
        self._eof = False

        self._archive = laffi.read_new()
        try:
            self._set_passwords(passwords or [])
            self._try_to_open(allowArchives=True)
        except ArchiveError as exception:
            if self.printDebug >= 2:
                print(f"[Info] Was not able to open {self._file} as given archive. Try to undo compressions next.")
                print(f"[Info] Exception: {exception}")
            try:
                laffi.read_free(self._archive)
                self._archive = laffi.read_new()

                self._set_passwords(passwords or [])
                self._try_to_open(allowArchives=False)
            except ArchiveError as exception2:
                raise exception2 from exception

    def format_name(self):
        return laffi.format_name(self._archive)

    def filter_names(self):
        allNames = [laffi.filter_name(self._archive, i) for i in range(laffi.filter_count(self._archive))]
        return [name for name in allNames if name != b'none']

    def _try_to_open(self, allowArchives: bool):
        laffi.get_read_filter_function('all')(self._archive)

        for formatToEnable in IterableArchive.ENABLED_FORMATS if allowArchives else ['raw']:
            try:
                laffi.get_read_format_function(formatToEnable)(self._archive)
            except ValueError as exception:
                # Ignore exceptions from formats that are not supported such as "rar5" in manylinux2014.
                if self.printDebug >= 3:
                    print(f"[Warning] Failed to enable format {formatToEnable} because of: {exception}")

        if isinstance(self._file, str):
            laffi.read_open_filename_w(self._archive, self._file, os.stat(self._file).st_blksize)
        elif isinstance(self._file, int):
            laffi.read_open_fd(self._archive, self._file, os.fstat(self._file).st_blksize)
        elif hasattr(self._file, 'readinto'):
            self._open_with_file_object()
        else:
            raise ValueError(
                "Libarchive backend currently only works with file path, descriptor, or a file object, "
                f"but got: {self._file}"
            )

        if not allowArchives and not self.filter_names():
            raise ArchiveError("When not looking for archives, there must be at least one filter!")

    def _set_passwords(self, passwords: List[Union[str, bytes]]):
        try:
            for password in passwords:
                laffi.read_add_passphrase(self._archive, password.encode() if isinstance(password, str) else password)
        except AttributeError as exception:
            raise NotImplementedError(
                f"This libarchive library version ({laffi.version_number()}) at "
                f"{laffi.libarchive_path}) does not support encryption!"
            ) from exception

    def _open_with_file_object(self):
        if not hasattr(self._file, 'readinto'):
            return
        self._buffer = ctypes.create_string_buffer(self.bufferSize)

        def read_from_file_object(_archive, _context, pointerToBufferPointer):
            try:
                size = self._file.readinto(self._buffer)
            except (NotImplementedError, AttributeError):
                result = self._file.read(self.bufferSize)
                size = len(result)
                self._buffer = ctypes.c_char_p(result)
            pointerToBufferPointer = ctypes.cast(pointerToBufferPointer, ctypes.POINTER(ctypes.c_void_p))
            pointerToBufferPointer[0] = ctypes.cast(self._buffer, ctypes.c_void_p)
            return size

        if hasattr(self._file, 'seekable') and self._file.seekable() and hasattr(self._file, 'tell'):

            def seek_file_object(_archive, _context, offset, whence):
                self._file.seek(offset, whence)
                return self._file.tell()

            self._file.seek(0)
            self._seekCallback = laffi.SEEK_CALLBACK(seek_file_object)
            # Needs to be assigned to self, or else, ctypes lifetime issues result in a segfault!
            laffi.read_set_seek_callback(self._archive, self._seekCallback)

        # Needs to be assigned to self, or else, ctypes lifetime issues result in a segfault!
        self._readCallback = laffi.READ_CALLBACK(read_from_file_object)
        laffi.read_open(self._archive, None, laffi.NO_OPEN_CB, self._readCallback, laffi.NO_CLOSE_CB)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        pass

    def __del__(self):
        laffi.read_free(self._archive)

    def next_entry(self) -> Optional[ArchiveEntry]:
        if self._eof:
            return None

        self._entry = ArchiveEntry(self._archive, entry_index=self._entryIndex, encoding=self.encoding)
        if self._entryIndex == 0 and self.printDebug >= 2:
            format_name = laffi.format_name(self._archive)
            if isinstance(format_name, bytes):
                format_name = format_name.decode()
            # We need to try and read the first entry before format_name returns anything other than 'none'.
            print(
                f"[Info] Successfully opened type '{format_name}' with libarchive. "
                f"Using filters: {self.filter_names()}"
            )

        self._entryIndex += 1
        if self._entry.eof:
            self._eof = True
            return None

        return self._entry

    def entry_index(self):
        return self._entryIndex

    def read_data(self, buffer, size):
        return laffi.read_data(self._archive, buffer, size)


class IterableArchiveCache:
    def __init__(self, cacheSize: int = 5):
        self.cacheSize = cacheSize
        self._cache: List[IterableArchive] = []

    def insert(self, archive: IterableArchive):
        self._cache.append(archive)
        while len(self._cache) > self.cacheSize:
            self._cache.pop(0)

    def take(self, entry_index: int) -> Optional[IterableArchive]:
        qualified = [i for i in range(len(self._cache)) if self._cache[i].entry_index() < entry_index]
        if not qualified:
            return None
        # Find the archive with the closest (highest) index not larger or equal than the requested one
        result = max(qualified, key=lambda i: self._cache[i].entry_index())
        return self._cache.pop(result)


class LibarchiveFile(io.RawIOBase):
    def __init__(
        self,
        file,
        entry_index,
        fileSize,
        passwords: Optional[List[str]] = None,
        printDebug: int = 0,
        archiveCache: Optional[IterableArchiveCache] = None,
    ):
        io.RawIOBase.__init__(self)
        self.file = file
        self.fileSize = fileSize
        self.entry_index = entry_index
        self.passwords = passwords
        self.printDebug = printDebug
        self._archiveCache = archiveCache
        self._archive = None
        self._entry = None
        self._bufferSizeMax = 1024 * 1024
        self._buffer = None
        self._bufferOffset = 0  # Offset in self.entry from which the buffer was read
        self._bufferIO = None

        self._open()

    def _open(self):
        self.close()

        self._archive = self._archiveCache.take(self.entry_index)
        if self._archive is None:
            self._archive = IterableArchive(self.file, passwords=self.passwords, printDebug=self.printDebug)

        while True:
            self._entry = self._archive.next_entry()
            if self._entry is None:
                break
            if self._entry.entry_index == self.entry_index:
                self._refill_buffer()
                break

        if self._entry is None and self._bufferIO is None:
            raise ValueError(f"Failed to find archive entry {self.entry_index}.")

    def _refill_buffer(self):
        bufferedSize = len(self._buffer) if self._buffer else 0
        if self._bufferOffset + bufferedSize >= self.fileSize:
            return

        self._bufferOffset += bufferedSize
        sizeToRead = min(self._bufferSizeMax, max(0, self.fileSize - self._bufferOffset))

        # Reallocate buffer if we need a different size. This is necessary because I don't see an API to
        # specify a size to io,BytesIO to make it work on a subset of the buffer without copying.
        if sizeToRead >= self._bufferSizeMax:
            if self._buffer is None or len(self._buffer) < self._bufferSizeMax:
                self._buffer = ctypes.create_string_buffer(self._bufferSizeMax)
        elif self._buffer is None or len(self._buffer) != sizeToRead:
            self._buffer = ctypes.create_string_buffer(sizeToRead)

        readSize = self._archive.read_data(self._buffer, sizeToRead)
        if readSize != sizeToRead:
            raise RuntimeError(f"Read {readSize} bytes but expected {self.fileSize} for entry {self.entry_index}!")
        self._bufferIO = io.BytesIO(self._buffer)

    @overrides(io.RawIOBase)
    def close(self) -> None:
        if self._archiveCache and self._archive:
            self._archiveCache.insert(self._archive)
        if self._bufferIO is not None:
            self._bufferIO.close()
            self._bufferIO = None
            self._bufferOffset = 0
        self._buffer = None
        self._entry = None
        self._archive = None

    @overrides(io.RawIOBase)
    def fileno(self) -> int:
        # This is a virtual Python level file object and therefore does not have a valid OS file descriptor!
        raise io.UnsupportedOperation

    @overrides(io.RawIOBase)
    def seekable(self) -> bool:
        return True

    @overrides(io.RawIOBase)
    def readable(self) -> bool:
        return self._bufferIO is not None and self._bufferIO.readable()

    @overrides(io.RawIOBase)
    def writable(self) -> bool:
        return False

    def read1(self, size: int = -1) -> bytes:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be read from!")
        result = self._bufferIO.read(size)
        if result:
            return result
        self._refill_buffer()
        return self._bufferIO.read(size)

    @overrides(io.RawIOBase)
    def read(self, size: int = -1) -> bytes:
        result = bytearray()
        while size < 0 or len(result) < size:
            read_data = self.read1(size if size < 0 else size - len(result))
            if not read_data:
                break
            result.extend(read_data)
        return bytes(result)

    def _skip(self, size: int) -> None:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be read from!")
        while size > 0:
            data = self.read1(size)
            if not data:
                break
            size -= len(data)

    @overrides(io.RawIOBase)
    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        if not self._bufferIO:
            raise RuntimeError("Closed file cannot be seeked!")

        here = self.tell()
        if whence == io.SEEK_CUR:
            offset += here
        elif whence == io.SEEK_END:
            offset += self.fileSize

        if offset >= here:
            self._skip(offset - here)
            return self.tell()

        bufferedSize = 0
        if self._buffer:
            bufferedSize = len(self._buffer)
        if offset >= self._bufferOffset and offset < self._bufferOffset + bufferedSize:
            self._bufferIO.seek(offset - self._bufferOffset)
        else:
            self._open()
            self._skip(max(0, offset))
        return self.tell()

    @overrides(io.RawIOBase)
    def tell(self) -> int:
        result = self._bufferOffset
        if self._bufferIO:
            result += self._bufferIO.tell()
        return result


# The implementation is similar to ZipMountSource and SQLiteIndexedTarUserData.
class LibarchiveMountSource(SQLiteIndexMountSource):
    # fmt: off
    def __init__(
        self,
        fileOrPath             : Union[str, IO[bytes]],
        writeIndex             : bool                      = False,
        clearIndexCache        : bool                      = False,
        indexFilePath          : Optional[str]             = None,
        indexFolders           : Optional[List[str]]       = None,
        encoding               : str                       = tarfile.ENCODING,
        verifyModificationTime : bool                      = False,
        printDebug             : int                       = 0,
        transform              : Optional[Tuple[str, str]] = None,
        indexMinimumFileCount  : int                       = 0,
        tarFileName            : Optional[str]             = None,
        **options
    ) -> None:
        self.archiveFilePath        = fileOrPath if isinstance(fileOrPath, str) else None
        self.fileOrPath             = fileOrPath
        self.encoding               = encoding
        self.verifyModificationTime = verifyModificationTime
        self.printDebug             = printDebug
        self.options                = options
        self.transformPattern       = transform
        self.passwords              = options.get("passwords", [])
        self.tarFileName            = tarFileName
        self._archiveCache          = IterableArchiveCache()
        # fmt: on

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        # Determine an archive file name to show for debug output and as file name inside the mount point for
        # simple non-TAR gzip/bzip2 stream-compressed files.
        if tarFileName:
            self.tarFileName = tarFileName
        elif isinstance(fileOrPath, str):
            self.tarFileName = fileOrPath
        else:
            self.tarFileName = '<file object>'

        # Force indexes in memory because:
        #  - They are incompatible with SQLiteIndexedTar and relevant consistency checks do not exist in older versions.
        #    Older versions would simply show the wrong folder hierarchy and return Input/Output error on file access.
        #  - Seeking to a file takes on average half as time much as creating the index. I.e., the overhead for
        #    creating the index feels relatively insignificant assuming that more than 2 files are accessed.
        indexFilePath = ':memory:'
        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=self.archiveFilePath,
                encoding=self.encoding,
                printDebug=self.printDebug,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='LibarchiveMountSource',
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        isFileObject = False  # Not supported yet

        if self.index.index_is_loaded():
            metadata = dict(self.index.get_connection().execute('SELECT * FROM metadata;'))
            if 'backend' not in metadata or metadata['backend'] != 'libarchive':
                self.__exit__(None, None, None)
                raise InvalidIndexError("The found index was not created by the libarchive backend.")

            self.index.reload_index_read_only()
        else:
            # Open new database when we didn't find an existing one.
            # Simply open in memory without an error even if writeIndex is True but when not indication
            # for a index file location has been given.
            if writeIndex and (indexFilePath or not isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

            self._create_index()
            if self.index.index_is_loaded():
                self._store_metadata()
                self.index.reload_index_read_only()

    def _create_index(self) -> None:
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()

        triedToOpen = False
        fileInfos = []
        gotAnyEntry = False
        with IterableArchive(self.fileOrPath, passwords=self.passwords, printDebug=self.printDebug) as archive:
            while True:
                entry = archive.next_entry()
                if entry is None:
                    break

                entryPath = None
                if entry.entry_index == 0 and self.tarFileName and entry.format_name() == b'raw':
                    libarchiveSuffixes = [
                        s for fid in COMPRESSION_BACKENDS['libarchive'].formats for s in FILE_FORMATS[fid].extensions
                    ]
                    fname = os.path.basename(self.tarFileName)
                    for suffix in ['gz', 'bz2', 'bzip2', 'gzip', 'xz', 'zst', 'zstd', *libarchiveSuffixes]:
                        suffix = '.' + suffix
                        if fname.lower().endswith(suffix.lower()) and len(fname) > len(suffix):
                            fname = fname[: -len(suffix)]
                            break
                    entryPath = fname

                gotAnyEntry = True
                fileInfos.append(entry.convert_to_row(entry.entry_index, self.transform, path=entryPath))
                # Contains file info SQLite row tuples! 4 -> size, 6 -> mode
                if not triedToOpen and not stat.S_ISDIR(fileInfos[-1][6]) and fileInfos[-1][4] > 0:
                    bufferSize = 1
                    buffer = ctypes.create_string_buffer(bufferSize)
                    try:
                        archive.read_data(buffer, 1)
                    except ArchiveError as exception:
                        # Very special case to delegate to py7zr somewhat smartly for encrypted 7z archives.
                        if entry.format_name() == b'7-Zip' and self.passwords and "py7zr" in sys.modules:
                            raise exception
                        if 'encrypt' in str(exception).lower() and self.printDebug >= 1:
                            print("[Warning] The file contents are encrypted but not the file hierarchy!")
                            print("[Warning] Specify a password with --password to also view file contents!")
                    triedToOpen = True

                if len(fileInfos) > 1000:
                    self.index.set_file_infos(fileInfos)
                    fileInfos = []

            if fileInfos:
                self.index.set_file_infos(fileInfos)
            elif not gotAnyEntry and archive.format_name() == b'tar':
                raise ArchiveError("Supposedly detected a TAR with no entries in it. Rejecting it as unknown format!")

        # Resort by (path,name). This one-time resort is faster than resorting on each INSERT (cache spill)
        if self.printDebug >= 2:
            print("Resorting files by path ...")

        self.index.finalize()

        t1 = timer()
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.archiveFilePath} took {t1 - t0:.2f}s")

    def _store_metadata(self) -> None:
        argumentsToSave = ['encoding', 'transformPattern']
        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        self.index.store_metadata(argumentsMetadata, self.archiveFilePath)
        self.index.store_metadata_key_value('backend', 'libarchive')

    def __del__(self):
        # TODO check that all objects are really closed to avoid memory leaks
        pass

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        assert fileInfo.userdata
        tarFileInfo = fileInfo.userdata[-1]
        assert isinstance(tarFileInfo, SQLiteIndexedTarUserData)
        return cast(
            IO[bytes],
            LibarchiveFile(
                self.fileOrPath,
                tarFileInfo.offsetheader,
                fileSize=fileInfo.size,
                passwords=self.passwords,
                printDebug=self.printDebug,
                archiveCache=self._archiveCache,
            ),
        )

    def _check_metadata(self, metadata: Dict[str, Any]) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""
        SQLiteIndex.check_archive_stats(self.archiveFilePath, metadata, self.verifyModificationTime)

        if 'arguments' in metadata:
            SQLiteIndex.check_metadata_arguments(
                json.loads(metadata['arguments']), self, argumentsToCheck=['encoding', 'transformPattern']
            )



#!/usr/bin/env python3

import site
import sys

from setuptools import setup

# workaround https://github.com/pypa/pip/issues/7953
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]
# Workaround for https://github.com/pypa/setuptools/issues/2938
#   Did not work in earlier pip versions: python3 -m pip install --user --editable .
#   Use: python3 setup.py develop --user .
# Workaround for https://github.com/pypa/setuptools/issues/3582#issuecomment-1242384337
#   python3 -m pip install --user --no-build-isolation --editable .
# https://github.com/pypa/pip/issues/6264#issuecomment-1184327415
# https://github.com/pypa/pip/pull/11466
# Upgrade to pip 22.3.1 or higher to avoid having to specify --no-build-isolation
setup()
